import pytest

from degenbot import (
    AnvilFork,
    Erc20Token,
    UniswapV2Pool,
    get_checksum_address,
    pool_registry,
    set_web3,
    token_registry,
)
from degenbot.exceptions import DegenbotValueError
from degenbot.exceptions.registry import RegistryAlreadyInitialized
from degenbot.registry.pool import PoolRegistry
from degenbot.registry.token import TokenRegistry

UNISWAP_V2_WBTC_WETH_POOL = get_checksum_address("0xBb2b8038a1640196FbE3e38816F3e67Cba72D940")
WETH_ADDRESS = get_checksum_address("0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2")


def test_singleton(fork_mainnet_full: AnvilFork):
    set_web3(fork_mainnet_full.w3)

    with pytest.raises(RegistryAlreadyInitialized):
        PoolRegistry()

    with pytest.raises(RegistryAlreadyInitialized):
        TokenRegistry()

    assert PoolRegistry.get_instance() is pool_registry
    assert TokenRegistry.get_instance() is token_registry


def test_adding_pool(fork_mainnet_full: AnvilFork):
    set_web3(fork_mainnet_full.w3)
    lp = UniswapV2Pool(UNISWAP_V2_WBTC_WETH_POOL)
    assert (
        pool_registry.get(pool_address=lp.address, chain_id=fork_mainnet_full.w3.eth.chain_id) is lp
    )

    with pytest.raises(DegenbotValueError):
        pool_registry.add(
            pool_address=lp.address, chain_id=fork_mainnet_full.w3.eth.chain_id, pool=lp
        )


def test_deleting_pool(fork_mainnet_full: AnvilFork):
    set_web3(fork_mainnet_full.w3)
    lp = UniswapV2Pool(UNISWAP_V2_WBTC_WETH_POOL)
    assert (
        pool_registry.get(pool_address=lp.address, chain_id=fork_mainnet_full.w3.eth.chain_id) is lp
    )
    pool_registry.remove(pool_address=lp.address, chain_id=fork_mainnet_full.w3.eth.chain_id)
    assert (
        pool_registry.get(pool_address=lp.address, chain_id=fork_mainnet_full.w3.eth.chain_id)
        is None
    )


def test_adding_token(fork_mainnet_full: AnvilFork):
    set_web3(fork_mainnet_full.w3)
    weth = Erc20Token(WETH_ADDRESS)
    assert (
        token_registry.get(token_address=weth.address, chain_id=fork_mainnet_full.w3.eth.chain_id)
        is weth
    )

    with pytest.raises(DegenbotValueError):
        token_registry.add(
            token_address=weth.address, chain_id=fork_mainnet_full.w3.eth.chain_id, token=weth
        )


def test_deleting_token(fork_mainnet_full: AnvilFork):
    set_web3(fork_mainnet_full.w3)
    weth = Erc20Token(WETH_ADDRESS)
    assert (
        token_registry.get(token_address=weth.address, chain_id=fork_mainnet_full.w3.eth.chain_id)
        is weth
    )
    token_registry.remove(token_address=weth.address, chain_id=fork_mainnet_full.w3.eth.chain_id)
    assert (
        token_registry.get(token_address=weth.address, chain_id=fork_mainnet_full.w3.eth.chain_id)
        is None
    )



from . import abi
from .managers import UniswapV2PoolManager, UniswapV3PoolManager
from .v2_liquidity_pool import UniswapV2Pool
from .v2_types import (
    UniswapV2PoolExternalUpdate,
    UniswapV2PoolSimulationResult,
    UniswapV2PoolState,
    UniswapV2PoolStateUpdated,
)
from .v3_liquidity_pool import UniswapV3Pool
from .v3_snapshot import UniswapV3LiquiditySnapshot
from .v3_types import (
    UniswapV3PoolExternalUpdate,
    UniswapV3PoolSimulationResult,
    UniswapV3PoolState,
    UniswapV3PoolStateUpdated,
)
from .v4_liquidity_pool import UniswapV4Pool
from .v4_snapshot import UniswapV4LiquiditySnapshot
from .v4_types import UniswapV4PoolExternalUpdate, UniswapV4PoolState, UniswapV4PoolStateUpdated

__all__ = (
    "UniswapV2Pool",
    "UniswapV2PoolExternalUpdate",
    "UniswapV2PoolManager",
    "UniswapV2PoolSimulationResult",
    "UniswapV2PoolState",
    "UniswapV2PoolStateUpdated",
    "UniswapV3LiquiditySnapshot",
    "UniswapV3Pool",
    "UniswapV3PoolExternalUpdate",
    "UniswapV3PoolManager",
    "UniswapV3PoolSimulationResult",
    "UniswapV3PoolState",
    "UniswapV3PoolStateUpdated",
    "UniswapV4LiquiditySnapshot",
    "UniswapV4Pool",
    "UniswapV4PoolExternalUpdate",
    "UniswapV4PoolState",
    "UniswapV4PoolStateUpdated",
    "abi",
)



import asyncio
from collections.abc import Awaitable, Iterable, Mapping, Sequence
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from fractions import Fraction
from typing import TYPE_CHECKING, Any
from weakref import WeakSet

import eth_abi.abi
from eth_typing import ChecksumAddress
from scipy.optimize import OptimizeResult, minimize_scalar
from web3 import Web3

from degenbot import connection_manager, get_checksum_address
from degenbot.arbitrage.types import (
    ArbitrageCalculationResult,
    CurveStableSwapPoolSwapAmounts,
    CurveStableSwapPoolVector,
    UniswapV2PoolSwapAmounts,
    UniswapV3PoolSwapAmounts,
)
from degenbot.constants import MAX_UINT256
from degenbot.curve.curve_stableswap_liquidity_pool import CurveStableswapPool
from degenbot.curve.types import CurveStableswapPoolState
from degenbot.erc20 import Erc20Token
from degenbot.exceptions import DegenbotValueError
from degenbot.exceptions.evm import EVMRevertError
from degenbot.exceptions.arbitrage import ArbitrageError, NoLiquidity
from degenbot.exceptions.liquidity_pool import LiquidityPoolError
from degenbot.logging import logger
from degenbot.types.abstract import AbstractArbitrage, AbstractLiquidityPool
from degenbot.types.aliases import BlockNumber
from degenbot.types.concrete import (
    AbstractPublisherMessage,
    PoolStateMessage,
    Publisher,
    PublisherMixin,
    Subscriber,
    TextMessage,
)
from degenbot.uniswap.types import UniswapPoolSwapVector
from degenbot.uniswap.v2_liquidity_pool import UniswapV2Pool
from degenbot.uniswap.v2_types import UniswapV2PoolState
from degenbot.uniswap.v3_libraries.tick_math import MAX_SQRT_RATIO, MIN_SQRT_RATIO
from degenbot.uniswap.v3_liquidity_pool import UniswapV3Pool
from degenbot.uniswap.v3_types import UniswapV3PoolState

type CurveOrUniswapPoolState = UniswapV2PoolState | UniswapV3PoolState | CurveStableswapPoolState
type CurveOrUniswapPool = UniswapV2Pool | UniswapV3Pool | CurveStableswapPool
type CurveOrUniswapSwapAmount = (
    CurveStableSwapPoolSwapAmounts | UniswapV2PoolSwapAmounts | UniswapV3PoolSwapAmounts
)


# Default discount applied to amount received.
# This masks small differences in get_dy() vs exchange().
CURVE_V1_DEFAULT_DISCOUNT_FACTOR = 0.9999


class UniswapCurveCycle(PublisherMixin, AbstractArbitrage):
    def _notify_subscribers(self: Publisher, message: AbstractPublisherMessage) -> None:
        for subscriber in self._subscribers:
            subscriber.notify(publisher=self, message=message)

    def __init__(
        self,
        input_token: Erc20Token,
        swap_pools: Iterable[CurveOrUniswapPool],
        id: str,  # noqa:A002
        max_input: int | None = None,
    ) -> None:
        for swap_pool in swap_pools:
            if not isinstance(swap_pool, CurveOrUniswapPool.__value__):
                raise DegenbotValueError(
                    message=f"Incompatible pool type ({type(swap_pool)}) provided."
                )

        self.swap_pools: tuple[CurveOrUniswapPool, ...] = tuple(swap_pools)
        self.name = " → ".join([pool.name for pool in self.swap_pools])

        self.id = id
        self.input_token = input_token
        self.curve_discount_factor = CURVE_V1_DEFAULT_DISCOUNT_FACTOR

        if max_input == 0:
            raise DegenbotValueError(message="Maximum input must be positive.")

        if max_input is None:
            logger.warning("No maximum input provided, setting to 100 WETH")
            max_input = 100 * 10**18
        self.max_input = max_input

        # Set up pre-determined "swap vectors", which allows the helper
        # to identify the tokens and direction of each swap along the path
        _swap_vectors: list[CurveStableSwapPoolVector | UniswapPoolSwapVector] = []
        for i, pool in enumerate(self.swap_pools):
            match pool:
                case UniswapV2Pool() | UniswapV3Pool():
                    if i == 0:
                        match self.input_token:
                            case pool.token0:
                                _swap_vectors.append(
                                    UniswapPoolSwapVector(
                                        token_in=pool.token0,
                                        token_out=pool.token1,
                                        zero_for_one=True,
                                    )
                                )
                            case pool.token1:
                                _swap_vectors.append(
                                    UniswapPoolSwapVector(
                                        token_in=pool.token1,
                                        token_out=pool.token0,
                                        zero_for_one=False,
                                    )
                                )
                            case _:  # pragma: no cover
                                raise DegenbotValueError(
                                    message="Input token could not be identified!"
                                )
                    else:
                        match _swap_vectors[-1].token_out:
                            case pool.token0:
                                _swap_vectors.append(
                                    UniswapPoolSwapVector(
                                        token_in=pool.token0,
                                        token_out=pool.token1,
                                        zero_for_one=True,
                                    )
                                )
                            case pool.token1:
                                _swap_vectors.append(
                                    UniswapPoolSwapVector(
                                        token_in=pool.token1,
                                        token_out=pool.token0,
                                        zero_for_one=False,
                                    )
                                )
                            case _:  # pragma: no cover
                                raise DegenbotValueError(
                                    message="Input token could not be identified!"
                                )
                case CurveStableswapPool():
                    # A Curve pool may have 3 or more tokens, so instead of a binary
                    # token0/token1 choice, determine the forward token by comparing
                    # current and next pool
                    if i != 1:
                        raise DegenbotValueError(
                            message="Not implemented for Curve pools at position != 1."
                        )

                    token_in = _swap_vectors[-1].token_out
                    next_pool = self.swap_pools[i + 1]
                    shared_tokens = list(
                        set(pool.tokens).intersection(next_pool.tokens),
                    )
                    assert len(shared_tokens) > 0, f"this: {pool.tokens}, next: {next_pool.tokens}"

                    # @dev this assumes the first shared token is the correct one to continue
                    token_out = shared_tokens[0]

                    _swap_vectors.append(
                        CurveStableSwapPoolVector(token_in=token_in, token_out=token_out)
                    )
                case _:  # pragma: no cover
                    raise DegenbotValueError(message="Pool type could not be identified")
        self._swap_vectors = tuple(_swap_vectors)

        self._subscribers: WeakSet[Subscriber] = WeakSet()
        for pool in self.swap_pools:
            pool.subscribe(self)

    def __getstate__(self) -> dict[str, Any]:
        # Remove objects that cannot be pickled and are unnecessary to perform
        # the calculation
        dropped_attributes = ("_subscribers",)

        return {key: value for key, value in self.__dict__.items() if key not in dropped_attributes}

    def __str__(self) -> str:
        return self.name

    def _build_swap_amounts(
        self,
        token_in_quantity: int,
        state_overrides: Mapping[ChecksumAddress, CurveOrUniswapPoolState],
        block_number: BlockNumber | None = None,
    ) -> tuple[CurveOrUniswapSwapAmount, ...]:
        """
        Generate inputs for all swaps along the arbitrage path, starting with the specified amount
        of the input token defined in the constructor.
        """

        _token_out_quantity = 0
        swap_amounts: list[CurveOrUniswapSwapAmount] = []
        for i, (pool, swap_vector) in enumerate(
            zip(self.swap_pools, self._swap_vectors, strict=True)
        ):
            pool_state_override = state_overrides.get(pool.address)
            _token_in_quantity = token_in_quantity if i == 0 else _token_out_quantity

            if _token_in_quantity == 0:
                raise ArbitrageError(message="Zero amount swap")

            try:
                match pool, pool_state_override, swap_vector:
                    case UniswapV2Pool(), UniswapV2PoolState() | None, UniswapPoolSwapVector():
                        _token_out_quantity = pool.calculate_tokens_out_from_tokens_in(
                            token_in=swap_vector.token_in,
                            token_in_quantity=_token_in_quantity,
                            override_state=pool_state_override,
                        )
                        if _token_out_quantity == 0:  # pragma: no cover
                            raise ArbitrageError(
                                message=f"Zero-output swap through pool {pool} @ {pool.address}"
                            )
                        swap_amounts.append(
                            UniswapV2PoolSwapAmounts(
                                pool=pool.address,
                                amounts_in=(_token_in_quantity, 0)
                                if swap_vector.zero_for_one
                                else (0, _token_in_quantity),
                                amounts_out=(0, _token_out_quantity)
                                if swap_vector.zero_for_one
                                else (_token_out_quantity, 0),
                            )
                        )
                    case UniswapV3Pool(), UniswapV3PoolState() | None, UniswapPoolSwapVector():
                        _token_out_quantity = pool.calculate_tokens_out_from_tokens_in(
                            token_in=swap_vector.token_in,
                            token_in_quantity=_token_in_quantity,
                            override_state=pool_state_override,
                        )
                        if _token_out_quantity == 0:  # pragma: no cover
                            raise ArbitrageError(
                                message=f"Zero-output swap through pool {pool} @ {pool.address}"
                            )
                        swap_amounts.append(
                            UniswapV3PoolSwapAmounts(
                                pool=pool.address,
                                amount_specified=_token_in_quantity,
                                zero_for_one=swap_vector.zero_for_one,
                                sqrt_price_limit_x96=MIN_SQRT_RATIO + 1
                                if swap_vector.zero_for_one
                                else MAX_SQRT_RATIO - 1,
                            )
                        )
                    case (
                        CurveStableswapPool(),
                        CurveStableswapPoolState() | None,
                        CurveStableSwapPoolVector(),
                    ):
                        _token_out_quantity = int(
                            self.curve_discount_factor
                            * pool.calculate_tokens_out_from_tokens_in(
                                token_in=swap_vector.token_in,
                                token_out=swap_vector.token_out,
                                token_in_quantity=_token_in_quantity,
                                override_state=pool_state_override,
                                block_identifier=block_number,
                            )
                        )
                        if _token_out_quantity == 0:  # pragma: no cover
                            raise ArbitrageError(
                                message=f"Zero-output swap through pool {pool} @ {pool.address}"
                            )
                        swap_amounts.append(
                            CurveStableSwapPoolSwapAmounts(
                                token_in=swap_vector.token_in,
                                token_in_index=pool.tokens.index(swap_vector.token_in),
                                token_out=swap_vector.token_out,
                                token_out_index=pool.tokens.index(swap_vector.token_out),
                                amount_in=_token_in_quantity,
                                min_amount_out=_token_out_quantity,
                                underlying=(
                                    pool.base_pool is not None
                                    and (
                                        swap_vector.token_in in pool.tokens_underlying
                                        or swap_vector.token_out in pool.tokens_underlying
                                    )
                                ),
                            ),
                        )
                    case _:  # pragma: no cover
                        raise DegenbotValueError(
                            message=f"Could not process pool {pool} and override {pool_state_override}"  # noqa:E501
                        )
            except LiquidityPoolError as exc:  # pragma: no cover
                raise ArbitrageError(message=str(exc)) from exc

        return tuple(swap_amounts)

    def _pre_calculation_check(
        self,
        state_overrides: Mapping[ChecksumAddress, CurveOrUniswapPoolState],
    ) -> None:
        """
        Perform liquidity and minimum rate of exchange checks and raise an exception if further
        optimization should be avoided.
        """

        # A scalar value representing the net amount of 1 input token across
        # the complete path (including fees).
        # profit_factor > 1.0 indicates a profitable trade.
        profit_factor: float = 1.0

        # Check each pool for liquidity in the direction of the trade and account for its current
        # price and fee. The prices are absolute (not decimal-corrected) since the decimals for
        # intermediate tokens cancel out.
        # e.g. for a WETH -> USDC -> USDT -> WETH arbitrage,
        # profit factor:
        #   [input: WETH] -> [pool0: USDC/WETH]
        #   * [pool1: USDT/USDC]
        #   * [pool2: WETH/USDT] == [output: WETH]
        for pool, vector in zip(self.swap_pools, self._swap_vectors, strict=True):
            pool_state = state_overrides.get(pool.address) or pool.state

            match pool, pool_state, vector:
                case UniswapV2Pool(), UniswapV2PoolState(), UniswapPoolSwapVector():
                    if pool_state.reserves_token0 == 0 or pool_state.reserves_token1 == 0:
                        raise NoLiquidity(message=f"V2 pool {pool.address} has no liquidity")
                    if pool_state.reserves_token1 == 1 and vector.zero_for_one:
                        raise NoLiquidity(
                            message=f"V2 pool {pool.address} has no liquidity for a 0 -> 1 swap"
                        )
                    if pool_state.reserves_token0 == 1 and not vector.zero_for_one:
                        raise NoLiquidity(
                            message=f"V2 pool {pool.address} has no liquidity for a 1 -> 0 swap"
                        )

                    price = pool_state.reserves_token1 / pool_state.reserves_token0
                    fee = pool.fee_token0 if vector.zero_for_one else pool.fee_token1
                    profit_factor *= (price if vector.zero_for_one else 1 / price) * (
                        (fee.denominator - fee.numerator) / fee.denominator
                    )
                case UniswapV3Pool(), UniswapV3PoolState(), UniswapPoolSwapVector():
                    if pool_state.sqrt_price_x96 == 0:
                        raise NoLiquidity(
                            message=f"V3 pool {pool.address} has no liquidity (not initialized)"
                        )
                    if pool_state.tick_bitmap == {}:
                        raise NoLiquidity(
                            message=f"V3 pool {pool.address} has no liquidity (empty bitmap)"
                        )
                    if pool_state.liquidity == 0:  # pragma: no cover
                        # Check if the swap is 0 -> 1 and has reached the lower limit of the price
                        # range
                        if pool_state.sqrt_price_x96 == MIN_SQRT_RATIO + 1 and vector.zero_for_one:
                            raise NoLiquidity(
                                message=f"V3 pool {pool.address} has no liquidity for a 0 -> 1 swap"
                            )
                        # Check if the swap is 1 -> 0 and has reached the upper limit of the price
                        # range
                        if (
                            pool_state.sqrt_price_x96 == MAX_SQRT_RATIO - 1
                            and not vector.zero_for_one
                        ):
                            raise NoLiquidity(
                                message=f"V3 pool {pool.address} has no liquidity for a 1 -> 0 swap"
                            )

                    price = (pool_state.sqrt_price_x96 * pool_state.sqrt_price_x96) / (
                        6277101735386680763835789423207666416102355444464034512896  # 2**192
                    )
                    # V3 fees are integer values representing hundredths of a bip (0.0001)
                    # e.g. fee=3000 represents 0.3%
                    fee = Fraction(pool.fee, 1000000)
                    profit_factor *= (price if vector.zero_for_one else 1 / price) * (
                        (fee.denominator - fee.numerator) / fee.denominator
                    )
                case CurveStableswapPool(), _, _:
                    price = 1.0 * (10**vector.token_out.decimals) / (10**vector.token_in.decimals)
                    fee = Fraction(pool.fee, pool.FEE_DENOMINATOR)
                    profit_factor *= price * ((fee.denominator - fee.numerator) / fee.denominator)
                case _:  # pragma: no cover
                    raise DegenbotValueError(
                        message=f"Could not process pool {pool}, state {pool_state}, and vector {vector}"  # noqa:E501
                    )

        if profit_factor < 1.0:
            raise ArbitrageError(
                message=f"No profitable arbitrage at current prices. Profit factor: {profit_factor}"
            )

    def _calculate(
        self,
        state_overrides: Mapping[ChecksumAddress, CurveOrUniswapPoolState],
        block_number: BlockNumber | None = None,
    ) -> ArbitrageCalculationResult:
        """
        Calculate the optimal arbitrage profit using the maximum input as an upper bound.
        """

        # bound the amount to be swapped
        bounds = (
            1.0,
            max(2.0, float(self.max_input)),
        )

        # bracket the initial guess for the algo
        bracket_amount = self.max_input
        bracket = (
            0.45 * bracket_amount,
            0.50 * bracket_amount,
            0.55 * bracket_amount,
        )

        def arb_profit(x: float) -> float:
            token_in_quantity = int(x)  # round the input down
            token_out_quantity: int = 0

            for i, (pool, swap_vector) in enumerate(
                zip(self.swap_pools, self._swap_vectors, strict=True)
            ):
                pool_override = state_overrides.get(pool.address)

                try:
                    match pool, pool_override:
                        case UniswapV2Pool(), UniswapV2PoolState() | None:
                            token_out_quantity = pool.calculate_tokens_out_from_tokens_in(
                                token_in=swap_vector.token_in,
                                token_in_quantity=token_in_quantity
                                if i == 0
                                else token_out_quantity,
                                override_state=pool_override,
                            )
                        case UniswapV3Pool(), UniswapV3PoolState() | None:
                            token_out_quantity = pool.calculate_tokens_out_from_tokens_in(
                                token_in=swap_vector.token_in,
                                token_in_quantity=token_in_quantity
                                if i == 0
                                else token_out_quantity,
                                override_state=pool_override,
                            )

                        case CurveStableswapPool(), CurveStableswapPoolState() | None:
                            token_out_quantity = int(
                                self.curve_discount_factor
                                * pool.calculate_tokens_out_from_tokens_in(
                                    token_in=swap_vector.token_in,
                                    token_in_quantity=(
                                        token_in_quantity if i == 0 else token_out_quantity
                                    ),
                                    token_out=swap_vector.token_out,
                                    override_state=pool_override,
                                    block_identifier=block_number,
                                )
                            )

                except (EVMRevertError, LiquidityPoolError):
                    # The optimizer might send invalid amounts into the swap calculation during
                    # iteration. We don't want it to stop, so catch the exception and pretend the
                    # swap resulted in zero output
                    token_out_quantity = 0
                    break

            # minimize_scalar requires the function to have a minimum value
            # for the solver to settle on an optimum input, so return the
            # negated profit
            return -float(token_out_quantity - token_in_quantity)

        opt: OptimizeResult = minimize_scalar(
            fun=arb_profit,
            method="bounded",
            bounds=bounds,
            bracket=bracket,
            options={"xatol": 1.0},
        )

        # Negate the result to convert to a sensible value (positive profit)
        best_profit = -int(opt.fun)
        swap_amount = int(opt.x)

        best_amounts = self._build_swap_amounts(
            token_in_quantity=swap_amount,
            state_overrides=state_overrides,
            block_number=block_number,
        )

        newest_state_block = None
        if not state_overrides:
            pool_state_blocks = tuple(
                block for pool in self.swap_pools if (block := pool.state.block) is not None
            )
            if len(pool_state_blocks) == len(self.swap_pools):
                newest_state_block = max(pool_state_blocks)

        return ArbitrageCalculationResult(
            id=self.id,
            input_token=self.input_token,
            profit_token=self.input_token,
            input_amount=swap_amount,
            profit_amount=best_profit,
            swap_amounts=best_amounts,
            state_block=newest_state_block,
        )

    def calculate(
        self,
        state_overrides: Mapping[ChecksumAddress, CurveOrUniswapPoolState] | None = None,
    ) -> ArbitrageCalculationResult:
        """
        Calculate the results of the arbitrage at the current pool states, or at one or more
        overridden pool states if provided.
        """

        if state_overrides is None:
            state_overrides = {}

        self._pre_calculation_check(state_overrides=state_overrides)

        return self._calculate(state_overrides=state_overrides)

    async def calculate_with_pool(
        self,
        executor: ProcessPoolExecutor | ThreadPoolExecutor,
        state_overrides: Mapping[ChecksumAddress, CurveOrUniswapPoolState] | None = None,
    ) -> Awaitable[Any]:
        """
        Wrap the arbitrage calculation into an asyncio future using the
        specified executor.

        Arguments
        ---------
        executor : Executor
            An executor (from `concurrent.futures`) to process the calculation
            work. Both `ThreadPoolExecutor` and `ProcessPoolExecutor` are
            supported, but `ProcessPoolExecutor` is recommended.
        state_overrides : Mapping[ChecksumAddress, StateOverrideTypes], optional
            A mapping (dict or dict-like) of pool states, keyed by the pool address.

        Returns
        -------
        A future which returns a `ArbitrageCalculationResult` (or exception)
        when awaited.

        Notes
        -----
        This is an async function that must be called with the `await` keyword.
        """

        if state_overrides is None:
            state_overrides = {}

        self._pre_calculation_check(state_overrides=state_overrides)

        if isinstance(executor, ProcessPoolExecutor) and any(
            pool.sparse_liquidity_map for pool in self.swap_pools if isinstance(pool, UniswapV3Pool)
        ):  # pragma: no cover
            raise DegenbotValueError(
                message=f"Cannot calculate {self} with executor. One or more V3 pools has a sparse liquidity map."  # noqa: E501
            )

        curve_pool = self.swap_pools[1]
        curve_swap_vector = self._swap_vectors[1]

        if TYPE_CHECKING:
            assert isinstance(curve_pool, CurveStableswapPool)
            assert isinstance(curve_swap_vector, CurveStableSwapPoolVector)

        block_number = connection_manager.get_web3(curve_pool.chain_id).eth.get_block_number()

        # Some Curve pools utilize on-chain lookups in their calc, so do a simple pre-calc to
        # cache those values for a given block since the pool will be disconnected once sent
        # into the process pool, e.g. it will have no web3 object for communication with the chain
        curve_pool.calculate_tokens_out_from_tokens_in(
            token_in=curve_swap_vector.token_in,
            token_in_quantity=1,
            token_out=curve_swap_vector.token_out,
            block_identifier=block_number,
        )

        return asyncio.get_running_loop().run_in_executor(
            executor,
            self._calculate,
            state_overrides,
            block_number,
        )

    def generate_payloads(
        self,
        *,
        from_address: ChecksumAddress | str,
        swap_amount: int,
        pool_swap_amounts: Sequence[CurveOrUniswapSwapAmount],
        infinite_approval: bool = False,
    ) -> list[tuple[ChecksumAddress, bytes, int]]:
        """
        Generate a list of tuple-formatted payloads for each step in the swap path.

        Calldata is built using the eth_abi.encode method and the ABI for the
        swap functions at each pool. Curve V1 and Uniswap V2/V3 pools are supported.

        Arguments
        ---------
        from_address: str
            The address that will execute the calldata. Must be a smart contract implementing the
            required callbacks specific to the pool.

        swap_amount: int
            The initial amount of `token_in` to swap through the first pool.

        pool_swap_amounts: Sequence[CurveOrUniswapSwapAmount]
            An ordered sequence of swap amounts.

        infinite_approval: bool
            Whether the infinite approval should be requested for the Curve pool. If False, only
            the minimum required amount will be approved.

        Returns
        -------
        ``list[(str, bytes, int)]``
            A list of payload tuples. Each payload tuple has form
            (
                address: ChecksumAddress,
                calldata: bytes,
                value: int
            ).
        """

        from_address = get_checksum_address(from_address)

        msg_value = 0  # This arbitrage does not require a `msg.value` payment
        payloads = []

        for i, (swap_pool, _swap_amounts) in enumerate(
            zip(self.swap_pools, pool_swap_amounts, strict=True)
        ):
            try:
                next_pool = self.swap_pools[i + 1]
            except IndexError:
                next_pool = None

            match next_pool:
                case UniswapV2Pool():
                    # V2 pools require a pre-swap transfer, so the contract does not have to
                    # perform intermediate custody and the swap can send the tokens directly to the
                    # next pool
                    swap_destination_address = next_pool.address
                case UniswapV3Pool() | CurveStableswapPool():
                    # V3 and Curve pools do not accept a pre-swap transfer, so the contract must
                    # maintain custody prior to a swap
                    swap_destination_address = from_address
                case None:
                    # Set the destination address for the last swap to the sending address
                    swap_destination_address = from_address
                case _:
                    raise DegenbotValueError(message=f"Unknown pool type {next_pool}")

            match i, swap_pool, _swap_amounts:
                case 0, UniswapV2Pool() as first_pool, _:
                    # Special case: If first pool is type V2, input token must be transferred prior
                    # to the swap
                    logger.debug(
                        f"PAYLOAD: transferring {swap_amount} WETH to V2 pool {first_pool}"
                    )
                    payloads.append(
                        (
                            # address
                            self.input_token.address,
                            # bytes calldata
                            Web3.keccak(text="transfer(address,uint256)")[:4]
                            + eth_abi.abi.encode(
                                types=(
                                    "address",
                                    "uint256",
                                ),
                                args=(
                                    first_pool.address,
                                    swap_amount,
                                ),
                            ),
                            msg_value,
                        )
                    )
                case _, UniswapV2Pool(), UniswapV2PoolSwapAmounts():
                    if _swap_amounts.amounts_out[0] == 0:
                        _token_in = swap_pool.token0
                        _token_out = swap_pool.token1
                        _amount_out = _swap_amounts.amounts_out[1]
                    else:
                        _token_in = swap_pool.token1
                        _token_out = swap_pool.token0
                        _amount_out = _swap_amounts.amounts_out[0]
                    logger.debug(f"PAYLOAD: building V2 swap at pool {i}")
                    logger.debug(f"PAYLOAD: pool address {swap_pool.address}")
                    logger.debug(f"PAYLOAD: swap {_token_in} -> {_amount_out} {_token_out} ")
                    logger.debug(f"PAYLOAD: destination address {swap_destination_address}")
                    payloads.append(
                        (
                            # address
                            swap_pool.address,
                            # bytes calldata
                            Web3.keccak(text="swap(uint256,uint256,address,bytes)")[:4]
                            + eth_abi.abi.encode(
                                types=(
                                    "uint256",
                                    "uint256",
                                    "address",
                                    "bytes",
                                ),
                                args=(
                                    *_swap_amounts.amounts_out,
                                    swap_destination_address,
                                    b"",
                                ),
                            ),
                            msg_value,
                        )
                    )
                case _, UniswapV3Pool(), UniswapV3PoolSwapAmounts():
                    if _swap_amounts.zero_for_one:
                        _token_in = swap_pool.token0
                        _token_out = swap_pool.token1
                    else:
                        _token_in = swap_pool.token1
                        _token_out = swap_pool.token0
                    logger.debug(f"PAYLOAD: building V3 swap at pool {i}")
                    logger.debug(f"PAYLOAD: pool address {swap_pool.address}")
                    logger.debug(f"PAYLOAD: token in  = {_token_in}")
                    logger.debug(f"PAYLOAD: token out = {_token_out}")
                    logger.debug(f"PAYLOAD: swap amounts {_swap_amounts}")
                    logger.debug(f"PAYLOAD: destination address {swap_destination_address}")
                    payloads.append(
                        (
                            # address
                            swap_pool.address,
                            # bytes calldata
                            Web3.keccak(text="swap(address,bool,int256,uint160,bytes)")[:4]
                            + eth_abi.abi.encode(
                                types=(
                                    "address",
                                    "bool",
                                    "int256",
                                    "uint160",
                                    "bytes",
                                ),
                                args=(
                                    swap_destination_address,
                                    _swap_amounts.zero_for_one,
                                    _swap_amounts.amount_specified,
                                    _swap_amounts.sqrt_price_limit_x96,
                                    b"",
                                ),
                            ),
                            msg_value,
                        )
                    )
                case _, CurveStableswapPool(), CurveStableSwapPoolSwapAmounts():
                    logger.debug(f"PAYLOAD: building Curve swap at pool {i}")
                    logger.debug(f"PAYLOAD: pool address {swap_pool.address}")
                    logger.debug(f"PAYLOAD: swap amounts {_swap_amounts}")
                    logger.debug(f"PAYLOAD: token in  = {_swap_amounts.token_in}")
                    logger.debug(f"PAYLOAD: token out = {_swap_amounts.token_out}")
                    logger.debug(f"PAYLOAD: destination address {swap_destination_address}")

                    current_approval = _swap_amounts.token_in.get_approval(
                        from_address, swap_pool.address
                    )
                    amount_to_approve: int | None = None
                    if infinite_approval is True and current_approval != MAX_UINT256:
                        amount_to_approve = MAX_UINT256
                    elif infinite_approval is False and current_approval < _swap_amounts.amount_in:
                        amount_to_approve = _swap_amounts.amount_in

                    if amount_to_approve is not None:
                        logger.debug(
                            f"PAYLOAD: approve {amount_to_approve} {_swap_amounts.token_in} by "
                            f"{swap_pool} {swap_destination_address}"
                        )
                        payloads.append(
                            (
                                # address
                                _swap_amounts.token_in.address,
                                # bytes calldata
                                Web3.keccak(text="approve(address,uint256)")[:4]
                                + eth_abi.abi.encode(
                                    types=["address", "uint256"],
                                    args=[swap_pool.address, amount_to_approve],
                                ),
                                msg_value,
                            )
                        )

                    logger.debug(
                        f"PAYLOAD: exchange {_swap_amounts.amount_in} {_swap_amounts.token_in_index}->{_swap_amounts.token_out_index}, min out = {_swap_amounts.min_amount_out}"  # noqa: E501
                    )
                    if _swap_amounts.underlying:
                        payloads.append(
                            (
                                # address
                                swap_pool.address,
                                # bytes calldata
                                Web3.keccak(
                                    text="exchange_underlying(int128,int128,uint256,uint256)"
                                )[:4]
                                + eth_abi.abi.encode(
                                    types=["int128", "int128", "uint256", "uint256"],
                                    args=[
                                        _swap_amounts.token_in_index,
                                        _swap_amounts.token_out_index,
                                        _swap_amounts.amount_in,
                                        _swap_amounts.min_amount_out,
                                    ],
                                ),
                                msg_value,
                            )
                        )
                    else:
                        payloads.append(
                            (
                                # address
                                swap_pool.address,
                                # bytes calldata
                                Web3.keccak(text="exchange(int128,int128,uint256,uint256)")[:4]
                                + eth_abi.abi.encode(
                                    types=["int128", "int128", "uint256", "uint256"],
                                    args=[
                                        _swap_amounts.token_in_index,
                                        _swap_amounts.token_out_index,
                                        _swap_amounts.amount_in,
                                        _swap_amounts.min_amount_out,
                                    ],
                                ),
                                msg_value,
                            )
                        )
                    if isinstance(next_pool, UniswapV2Pool):
                        logger.debug(
                            f"PAYLOAD: transferring {_swap_amounts.min_amount_out} {_swap_amounts.token_out} to V2 pool {next_pool}"  # noqa: E501
                        )
                        payloads.append(
                            (
                                # address
                                _swap_amounts.token_out.address,
                                # bytes calldata
                                Web3.keccak(text="transfer(address,uint256)")[:4]
                                + eth_abi.abi.encode(
                                    types=(
                                        "address",
                                        "uint256",
                                    ),
                                    args=(
                                        next_pool.address,
                                        _swap_amounts.min_amount_out,
                                    ),
                                ),
                                msg_value,
                            )
                        )
                case _:  # pragma: no cover
                    raise DegenbotValueError(
                        message=f"Could not identify pool {swap_pool} and amounts {_swap_amounts}"
                    )

        return payloads

    def notify(self, publisher: Publisher, message: AbstractPublisherMessage) -> None:
        match publisher, message:
            case (
                AbstractLiquidityPool(),
                PoolStateMessage(),
            ) if publisher in self.swap_pools:
                self._notify_subscribers(TextMessage(f"Received pool update from {publisher}"))
            case _:  # pragma: no cover
                logger.info(f"Unhandled message {message} from publisher {publisher}")



import dataclasses

import pydantic

from degenbot.types.abstract import AbstractPoolState, AbstractSimulationResult
from degenbot.types.aliases import BlockNumber
from degenbot.types.concrete import PoolStateMessage
from degenbot.validation.evm_values import ValidatedInt128, ValidatedUint128, ValidatedUint256

type BitmapWord = int
type Pip = int  # V3 pool fees are expressed in pips equaling one hundredth of 1%
type Liquidity = int
type LiquidityGross = int
type LiquidityNet = int
type SqrtPriceX96 = int
type Tick = int
type TickBitmap = int


@dataclasses.dataclass(slots=True, frozen=True)
class UniswapSimulationResult(AbstractSimulationResult):
    """
    Common attributes for Uniswap V2 & V3 simulations
    """

    amount0_delta: int
    amount1_delta: int
    initial_state: AbstractPoolState
    final_state: AbstractPoolState


class UniswapV3BitmapAtWord(pydantic.BaseModel, frozen=True):
    bitmap: ValidatedUint256
    block: BlockNumber = 0


class UniswapV3LiquidityAtTick(pydantic.BaseModel, frozen=True):
    liquidity_net: ValidatedInt128
    liquidity_gross: ValidatedUint128
    block: BlockNumber = 0


@dataclasses.dataclass(slots=True)
class UniswapV3LiquidityEvent:
    block_number: BlockNumber
    liquidity: Liquidity
    tick_lower: Tick
    tick_upper: Tick
    tx_index: int
    log_index: int


@dataclasses.dataclass(slots=True, frozen=True, eq=False)
class UniswapV3PoolExternalUpdate:
    block_number: BlockNumber
    liquidity: Liquidity
    sqrt_price_x96: SqrtPriceX96
    tick: Tick


@dataclasses.dataclass(slots=True, frozen=True, eq=False)
class UniswapV3PoolLiquidityMappingUpdate:
    block_number: BlockNumber
    liquidity: Liquidity
    tick_lower: Tick
    tick_upper: Tick


@dataclasses.dataclass(slots=True, frozen=True, kw_only=True)
class UniswapV3PoolState(AbstractPoolState):
    liquidity: Liquidity
    sqrt_price_x96: SqrtPriceX96
    tick: Tick
    tick_bitmap: dict[BitmapWord, UniswapV3BitmapAtWord]
    tick_data: dict[Tick, UniswapV3LiquidityAtTick]


@dataclasses.dataclass(slots=True, frozen=True)
class UniswapV3PoolSimulationResult(UniswapSimulationResult):
    initial_state: UniswapV3PoolState
    final_state: UniswapV3PoolState


@dataclasses.dataclass(slots=True, frozen=True)
class UniswapV3PoolStateUpdated(PoolStateMessage):
    state: UniswapV3PoolState


type InitializedTickMap = dict[BitmapWord, UniswapV3BitmapAtWord]
type LiquidityMap = dict[Tick, UniswapV3LiquidityAtTick]



import pathlib
from typing import Annotated, ClassVar

import pydantic
from sqlalchemy import Dialect, ForeignKey, Index, String, Text, create_engine
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship, sessionmaker
from sqlalchemy.types import TypeDecorator

from degenbot.config import settings
from degenbot.logging import logger

type Tick = int
type Word = int


class TicksAtWord(pydantic.BaseModel):
    bitmap: int


class LiquidityAtTick(pydantic.BaseModel):
    liquidity_net: int
    liquidity_gross: int


class PoolLiquidityMap(pydantic.BaseModel):
    tick_bitmap: dict[Word, TicksAtWord]
    tick_data: dict[Tick, LiquidityAtTick]


class IntMappedToString(TypeDecorator[int]):
    """
    EVM integers can be up to 32 bytes, which exceeds the usual 8 byte limit for most SQL backends.
    Map these values to a 78 character VARCHAR which can hold a string representation of all
    possible values.
    """

    cache_ok = True
    impl = String(78)

    def process_bind_param(
        self,
        value: int | None,
        dialect: Dialect,  # noqa: ARG002
    ) -> str | None:
        """
        Perform the Python type -> DB type conversion.
        """

        assert isinstance(value, int | None)
        return None if value is None else str(value)

    def process_result_value(
        self,
        value: str | None,
        dialect: Dialect,  # noqa: ARG002
    ) -> int | None:
        """
        Perform the DB type -> Python type conversion.
        """

        assert isinstance(value, str | None)
        return None if value is None else int(value)


Address = Annotated[str, mapped_column(String(42))]
BigInteger = Annotated[int, IntMappedToString]
PrimaryKeyInteger = Annotated[int, mapped_column(primary_key=True)]
IndexedForeignKeyPoolId = Annotated[int, mapped_column(ForeignKey("pools.id"), index=True)]


class Base(DeclarativeBase):
    type_annotation_map: ClassVar = {
        # keys must be Python types (native or Annotated)
        # values must be SQLAlchemy types
        BigInteger: IntMappedToString,
        str: Text,
    }


class Erc20TokenTableEntry(Base):
    __tablename__ = "erc20_tokens"

    id: Mapped[PrimaryKeyInteger]
    address: Mapped[Address]
    chain: Mapped[int]
    name: Mapped[str]
    symbol: Mapped[str]
    decimals: Mapped[int]


# A (address, ChainId) tuple is unique for ERC-20 tokens
Index(
    "ix_erc20_tokens_address_chain",
    Erc20TokenTableEntry.address,
    Erc20TokenTableEntry.chain,
    unique=True,
)


class MetadataTableEntry(Base):
    __tablename__ = "metadata"

    id: Mapped[PrimaryKeyInteger]
    key: Mapped[str]
    value: Mapped[str]


class LiquidityPositionTableEntry(Base):
    __tablename__ = "liquidity_positions"

    id: Mapped[PrimaryKeyInteger]
    pool_id: Mapped[IndexedForeignKeyPoolId]
    tick: Mapped[int]
    liquidity_net: Mapped[BigInteger]
    liquidity_gross: Mapped[BigInteger]


# A (PoolId, tick) tuple is unique for each liquidity position
Index(
    "ix_liquidity_positions_pool_id_tick",
    LiquidityPositionTableEntry.pool_id,
    LiquidityPositionTableEntry.tick,
    unique=True,
)


class InitializationMapTableEntry(Base):
    __tablename__ = "initialization_maps"

    id: Mapped[PrimaryKeyInteger]
    pool_id: Mapped[IndexedForeignKeyPoolId]
    word: Mapped[int]
    bitmap: Mapped[BigInteger]


# A (PoolId, word) tuple is unique for each initialization map
Index(
    "ix_initialization_maps_pool_id_word",
    InitializationMapTableEntry.pool_id,
    InitializationMapTableEntry.word,
    unique=True,
)


class Pool(Base):
    __tablename__ = "pools"

    id: Mapped[PrimaryKeyInteger]
    address: Mapped[Address] = mapped_column(unique=True, index=True)
    chain: Mapped[int]
    kind: Mapped[str] = mapped_column()

    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_on": kind,
        "polymorphic_identity": "pool",
    }


class AbstractUniswapPool(Pool):
    """
    This abstract class serves as a parent container for columns common to all Uniswap variant
    child classes. This class should not be directly instantiated.
    """

    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_abstract": True,
    }

    token0: Mapped[Address]
    token1: Mapped[Address]
    factory: Mapped[Address | None]
    deployer: Mapped[Address | None]


class UniswapV2Pool(AbstractUniswapPool):
    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_identity": "uniswap_v2",
    }


class UniswapV3Pool(AbstractUniswapPool):
    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_identity": "uniswap_v3",
    }

    fee: Mapped[int]
    tick_spacing: Mapped[int]
    has_liquidity: Mapped[bool]

    liquidity_positions: Mapped[list[LiquidityPositionTableEntry] | None] = relationship()
    initialization_maps: Mapped[list[InitializationMapTableEntry] | None] = relationship()


class AerodromeV3Pool(UniswapV3Pool):
    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_identity": "aerodrome_v3",
    }


class PancakeswapV3Pool(UniswapV3Pool):
    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_identity": "pancakeswap_v3",
    }


class SushiswapV3Pool(UniswapV3Pool):
    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_identity": "sushiswap_v3",
    }


def create_new_sqlite_database(db_path: pathlib.Path) -> None:
    if db_path.exists():
        db_path.unlink()

    engine = create_engine(
        f"sqlite:///{db_path.absolute()}",
    )
    Base.metadata.create_all(bind=engine)
    logger.info(f"Initialized new SQLite database at {db_path}")


def upgrade_existing_sqlite_database(db_path: pathlib.Path) -> None:
    db_path_abs = db_path.absolute()
    engine = create_engine(
        f"sqlite:///{db_path_abs}",
    )
    Base.metadata.create_all(bind=engine)
    logger.info(f"Updated existing SQLite database at {db_path_abs}")


_default_engine = create_engine(
    f"sqlite:///{settings.database.path.absolute()}",
)
default_session = sessionmaker(_default_engine)()



import itertools
from collections.abc import Iterable
from fractions import Fraction
from typing import TYPE_CHECKING

import eth_abi.packed
from eth_typing import ChecksumAddress
from eth_utils.crypto import keccak
from hexbytes import HexBytes

from degenbot.functions import create2_address

if TYPE_CHECKING:
    from degenbot.uniswap.managers import UniswapV2PoolManager
    from degenbot.uniswap.v2_liquidity_pool import UniswapV2Pool


def generate_v2_pool_address(
    deployer_address: str | bytes,
    token_addresses: Iterable[str | bytes],
    init_hash: str | bytes,
) -> ChecksumAddress:
    """
    Get the deterministic V2 pool address generated by CREATE2. Uses the token address to generate
    the salt. The token addresses can be passed in any order.

    Adapted from https://github.com/Uniswap/universal-router/blob/59f1291d3760d2537a7bd1cbf37317922a49efb0/contracts/modules/uniswap/v2/UniswapV2Library.sol#L50
    """

    sorted_token_addresses = sorted([HexBytes(address) for address in token_addresses])

    salt = keccak(
        eth_abi.packed.encode_packed(
            ("address", "address"),
            sorted_token_addresses,
        )
    )

    return create2_address(
        deployer=deployer_address,
        salt=salt,
        init_code_hash=init_hash,
    )


def get_v2_pools_from_token_path(
    tx_path: Iterable[ChecksumAddress | str],
    pool_manager: "UniswapV2PoolManager",
) -> list["UniswapV2Pool"]:
    result: list[UniswapV2Pool] = []
    for token_addresses in itertools.pairwise(tx_path):
        pool = pool_manager.get_pool_from_tokens(
            token_addresses=token_addresses,
            silent=True,
        )
        if TYPE_CHECKING:
            assert isinstance(pool, UniswapV2Pool)
        result.append(pool)
    return result


def constant_product_calc_exact_in(
    amount_in: int,
    reserves_in: int,
    reserves_out: int,
    fee: Fraction,
) -> int:
    """
    Calculate the amount out for an exact input from a constant product (x*y=k) invariant pool.
    """

    return (amount_in * (fee.denominator - fee.numerator) * reserves_out) // (
        reserves_in * fee.denominator + amount_in * (fee.denominator - fee.numerator)
    )


def constant_product_calc_exact_out(
    amount_out: int,
    reserves_in: int,
    reserves_out: int,
    fee: Fraction,
) -> int:
    """
    Calculate the amount in necessary for an exact output swap through a constant product (x*y=k)
    invariant pool.
    """

    return 1 + (reserves_in * amount_out * fee.denominator) // (
        (reserves_out - amount_out) * (fee.denominator - fee.numerator)
    )



import dataclasses

from eth_typing import ChecksumAddress
from hexbytes import HexBytes

from degenbot.types.abstract import AbstractPoolState
from degenbot.types.aliases import BlockNumber
from degenbot.types.concrete import PoolStateMessage
from degenbot.uniswap.v3_types import (
    BitmapWord,
    Liquidity,
    Pip,
    SqrtPriceX96,
    Tick,
    UniswapV3BitmapAtWord,
    UniswapV3LiquidityAtTick,
    UniswapV3LiquidityEvent,
    UniswapV3PoolExternalUpdate,
    UniswapV3PoolLiquidityMappingUpdate,
)

type FeeToProtocol = int
type SwapFee = int


class UniswapV4BitmapAtWord(UniswapV3BitmapAtWord): ...


class UniswapV4LiquidityAtTick(UniswapV3LiquidityAtTick): ...


@dataclasses.dataclass(slots=True, frozen=True, kw_only=True)
class UniswapV4PoolState(AbstractPoolState):
    liquidity: Liquidity
    sqrt_price_x96: SqrtPriceX96
    tick: Tick
    tick_bitmap: dict[BitmapWord, UniswapV4BitmapAtWord]
    tick_data: dict[Tick, UniswapV4LiquidityAtTick]
    id: HexBytes
    block: BlockNumber | None


@dataclasses.dataclass(slots=True, frozen=True)
class UniswapV4PoolKey:
    currency0: ChecksumAddress
    currency1: ChecksumAddress
    fee: Pip
    tick_spacing: int
    hooks: ChecksumAddress


class UniswapV4LiquidityEvent(UniswapV3LiquidityEvent): ...


class UniswapV4PoolExternalUpdate(UniswapV3PoolExternalUpdate): ...


class UniswapV4PoolLiquidityMappingUpdate(UniswapV3PoolLiquidityMappingUpdate): ...


@dataclasses.dataclass(slots=True, frozen=True)
class UniswapV4PoolStateUpdated(PoolStateMessage):
    state: UniswapV4PoolState


type InitializedTickMap = dict[BitmapWord, UniswapV4BitmapAtWord]
type LiquidityMap = dict[Tick, UniswapV4LiquidityAtTick]



from degenbot.uniswap.abi import UNISWAP_V2_POOL_ABI as SUSHISWAP_V2_POOL_ABI
from degenbot.uniswap.abi import UNISWAP_V3_POOL_ABI as SUSHISWAP_V3_POOL_ABI

__all__ = (
    "SUSHISWAP_V2_POOL_ABI",
    "SUSHISWAP_V3_POOL_ABI",
)




from pydantic import validate_call

from degenbot.validation.evm_values import ValidatedUint256


@validate_call(validate_return=True)
def div_rounding_up(
    x: ValidatedUint256,
    y: ValidatedUint256,
) -> ValidatedUint256:
    """
    Calculates ceil(x/y)

    @dev division by 0 will return 0, and should be checked externally.
    """

    return 0 if y == 0 else x // y + int(x % y > 0)


@validate_call(validate_return=True)
def simple_mul_div(
    a: ValidatedUint256,
    b: ValidatedUint256,
    denominator: ValidatedUint256,
) -> ValidatedUint256:
    """
    Calculates floor((a*b)/denominator))

    @dev division by 0 will return 0, and should be checked externally
    """

    return 0 if denominator == 0 else (a * b) // denominator



from .arbitrage import AbstractArbitrage
from .deployment import AbstractExchangeDeployment
from .erc20_token import AbstractErc20Token
from .liquidity_pool import AbstractLiquidityPool
from .pool_manager import AbstractPoolManager
from .pool_state import AbstractPoolState


class AbstractSimulationResult: ...


class AbstractPoolUpdate: ...


class AbstractManager:
    """
    Base class for managers that generate, track and distribute various helper classes
    """


class AbstractRegistry: ...


class AbstractTransaction: ...


__all__ = (
    "AbstractArbitrage",
    "AbstractErc20Token",
    "AbstractExchangeDeployment",
    "AbstractLiquidityPool",
    "AbstractManager",
    "AbstractPoolManager",
    "AbstractPoolState",
    "AbstractPoolUpdate",
    "AbstractRegistry",
    "AbstractSimulationResult",
    "AbstractTransaction",
)



from eth_typing import ChecksumAddress
from hexbytes import HexBytes


class AbstractErc20Token:
    address: ChecksumAddress
    symbol: str
    name: str
    decimals: int

    def __eq__(self, other: object) -> bool:
        match other:
            case AbstractErc20Token():
                return self.address == other.address
            case HexBytes():
                return self.address.lower() == other.to_0x_hex().lower()
            case bytes():
                return self.address.lower() == "0x" + other.hex().lower()
            case str():
                return self.address.lower() == other.lower()
            case _:
                return NotImplemented

    def __lt__(self, other: object) -> bool:
        match other:
            case AbstractErc20Token():
                return self.address < other.address
            case HexBytes():
                return self.address.lower() < other.to_0x_hex().lower()
            case bytes():
                return self.address.lower() < "0x" + other.hex().lower()
            case str():
                return self.address.lower() < other.lower()
            case _:
                return NotImplemented

    def __gt__(self, other: object) -> bool:
        match other:
            case AbstractErc20Token():
                return self.address > other.address
            case HexBytes():
                return self.address.lower() > other.to_0x_hex().lower()
            case bytes():
                return self.address.lower() > "0x" + other.hex().lower()
            case str():
                return self.address.lower() > other.lower()
            case _:
                return NotImplemented

    def __hash__(self) -> int:
        return hash(self.address)

    def __str__(self) -> str:
        return self.symbol



from degenbot.constants import MAX_UINT256, MIN_UINT256
from degenbot.exceptions.evm import EVMRevertError
from degenbot.uniswap.v3_libraries.functions import mulmod


def muldiv(
    a: int,
    b: int,
    denominator: int,
) -> int:
    """
    The Solidity implementation is designed to calculate a * b / d without risk of overflowing
    the intermediate result.

    Python integers do not overflow and have no bit depth limitation, so this function simply
    checks for an invalid result.

    ref: https://github.com/Uniswap/v3-core/blob/main/contracts/libraries/FullMath.sol
    """

    # Assert values are valid for Solidity contract
    if a < MIN_UINT256 or a > MAX_UINT256:
        raise EVMRevertError(error="Invalid value for a.")
    if b < MIN_UINT256 or b > MAX_UINT256:
        raise EVMRevertError(error="Invalid value for b.")
    if denominator < MIN_UINT256 or denominator > MAX_UINT256:
        raise EVMRevertError(error="Invalid value for denominator.")

    if denominator == 0:
        raise EVMRevertError(error="DIVISION BY ZERO")

    result = (a * b) // denominator

    if not (MIN_UINT256 <= result <= MAX_UINT256):
        raise EVMRevertError(error="Invalid result, does not fit in uint256")

    return result


def muldiv_rounding_up(a: int, b: int, denominator: int) -> int:
    result = muldiv(a, b, denominator)
    if mulmod(a, b, denominator) > 0:
        # must be less than max uint256 since we're rounding up
        if not (MIN_UINT256 <= result < MAX_UINT256):
            raise EVMRevertError(error="FAIL!")
        return result + 1
    return result



def div_rounding_up(x: int, y: int) -> int:
    """
    Perform an x//y floored division, rounding up any remainder.

    ref: https://github.com/Uniswap/v3-core/blob/main/contracts/libraries/UnsafeMath.sol
    """

    # x and y are uint256 values, so negative value floor division workarounds are unnecessary
    return (0 if y == 0 else x // y) + ((0 if y == 0 else x % y) > 0)



import contextlib
from typing import TYPE_CHECKING, cast

import eth_abi.abi
from eth_abi.exceptions import DecodingError
from eth_typing import ChecksumAddress
from sqlalchemy import select
from sqlalchemy.orm import Session
from web3 import AsyncWeb3, Web3
from web3.exceptions import Web3Exception
from web3.types import BlockIdentifier

import degenbot.registry
from degenbot import get_checksum_address
from degenbot.chainlink import ChainlinkPriceContract
from degenbot.connection import async_connection_manager, connection_manager
from degenbot.database import Erc20TokenTableEntry, default_session
from degenbot.exceptions import DegenbotValueError
from degenbot.exceptions.erc20 import NoPriceOracle
from degenbot.functions import (
    encode_function_calldata,
    get_number_for_block_identifier,
    get_number_for_block_identifier_async,
    raw_call,
)
from degenbot.logging import logger
from degenbot.types.abstract import AbstractErc20Token
from degenbot.types.aliases import BlockNumber, ChainId
from degenbot.types.concrete import BoundedCache

if TYPE_CHECKING:
    from hexbytes import HexBytes


def add_token_to_database(
    token: Erc20TokenTableEntry,
    session: Session = default_session,
) -> None:
    with session.begin():
        session.add(token)


def get_token_from_database(
    token: ChecksumAddress,
    chain_id: int,
    session: Session = default_session,
) -> Erc20TokenTableEntry | None:
    with session:
        return session.scalar(
            select(Erc20TokenTableEntry).where(
                Erc20TokenTableEntry.address == token,
                Erc20TokenTableEntry.chain == chain_id,
            )
        )


class Erc20Token(AbstractErc20Token):
    """
    An ERC-20 token contract.
    """

    UNKNOWN_NAME = "Unknown"
    UNKNOWN_SYMBOL = "UNKN"
    UNKNOWN_DECIMALS = 18

    def __init__(
        self,
        address: str,
        *,
        chain_id: ChainId | None = None,
        oracle_address: str | None = None,
        silent: bool = False,
        state_cache_depth: int = 8,
        use_database: bool = True,
    ) -> None:
        self.address = get_checksum_address(address)
        self._chain_id = chain_id if chain_id is not None else connection_manager.default_chain_id

        token_from_db: Erc20TokenTableEntry | None = None
        if use_database:
            token_from_db = get_token_from_database(
                token=self.address,
                chain_id=self.chain_id,
            )

        if token_from_db:
            self.name = token_from_db.name
            self.symbol = token_from_db.symbol
            self.decimals = token_from_db.decimals
        else:
            self.name = self.UNKNOWN_NAME
            self.symbol = self.UNKNOWN_SYMBOL
            self.decimals = self.UNKNOWN_DECIMALS

            w3 = self.w3
            try:
                self.name, self.symbol, self.decimals = self.get_name_symbol_decimals_batched(w3=w3)
            except (Web3Exception, DecodingError):
                for func_prototype in ("name()", "NAME()"):
                    try:
                        self.name = self.get_name(w3=w3, func_prototype=func_prototype)
                    except (Web3Exception, DecodingError):
                        continue
                    break

                for func_prototype in ("symbol()", "SYMBOL()"):
                    try:
                        self.symbol = self.get_symbol(w3=w3, func_prototype=func_prototype)
                    except (Web3Exception, DecodingError):
                        continue
                    break

                for func_prototype in ("decimals()", "DECIMALS()"):
                    try:
                        self.decimals = self.get_decimals(w3=w3, func_prototype=func_prototype)
                    except (Web3Exception, DecodingError):
                        continue
                    break

            if all(
                [
                    self.name == self.UNKNOWN_NAME,
                    self.symbol == self.UNKNOWN_SYMBOL,
                    self.decimals == self.UNKNOWN_DECIMALS,
                ]
            ) and not w3.eth.get_code(self.address):
                raise DegenbotValueError(message="No contract deployed at this address")

        self._price_oracle = (
            ChainlinkPriceContract(address=oracle_address, chain_id=self.chain_id)
            if oracle_address
            else None
        )

        degenbot.registry.token_registry.add(
            token_address=self.address, chain_id=self.chain_id, token=self
        )

        self._state_cache_depth = state_cache_depth
        self._cached_approval: dict[tuple[int, ChecksumAddress, ChecksumAddress], int] = {}
        self._cached_balance: dict[ChecksumAddress, BoundedCache[BlockNumber, int]] = {}
        self._cached_total_supply: BoundedCache[BlockNumber, int] = BoundedCache(
            max_items=state_cache_depth,
        )

        if not silent:  # pragma: no cover
            logger.info(f"• {self.symbol} ({self.name})")

        if use_database and token_from_db is None:
            add_token_to_database(
                Erc20TokenTableEntry(
                    address=self.address,
                    chain=self.chain_id,
                    name=self.name,
                    symbol=self.symbol,
                    decimals=self.decimals,
                )
            )

    def __repr__(self) -> str:  # pragma: no cover
        return f"{self.__class__.__name__}(address={self.address}, symbol='{self.symbol}', name='{self.name}', decimals={self.decimals})"  # noqa:E501

    def get_name_symbol_decimals_batched(self, w3: Web3) -> tuple[str, str, int]:
        with w3.batch_requests() as batch:
            batch.add_mapping(
                {
                    w3.eth.call: [
                        {
                            "to": self.address,
                            "data": encode_function_calldata(
                                function_prototype="name()",
                                function_arguments=None,
                            ),
                        },
                        {
                            "to": self.address,
                            "data": encode_function_calldata(
                                function_prototype="symbol()",
                                function_arguments=None,
                            ),
                        },
                        {
                            "to": self.address,
                            "data": encode_function_calldata(
                                function_prototype="decimals()",
                                function_arguments=None,
                            ),
                        },
                    ]
                }
            )

            name, symbol, decimals = batch.execute()

            (name,) = eth_abi.abi.decode(types=["string"], data=cast("HexBytes", name))
            (symbol,) = eth_abi.abi.decode(types=["string"], data=cast("HexBytes", symbol))
            (decimals,) = eth_abi.abi.decode(types=["uint256"], data=cast("HexBytes", decimals))

            return cast("str", name), cast("str", symbol), cast("int", decimals)

    def get_name(self, w3: Web3, func_prototype: str) -> str:
        result = w3.eth.call(
            {
                "to": self.address,
                "data": encode_function_calldata(
                    function_prototype=func_prototype,
                    function_arguments=None,
                ),
            }
        )

        try:
            (name,) = eth_abi.abi.decode(types=["string"], data=result)
            return cast("str", name)
        except DecodingError:
            (name,) = eth_abi.abi.decode(types=["bytes32"], data=result)
            return cast("HexBytes", name).decode("utf-8", errors="ignore").strip("\x00")

    def get_symbol(self, w3: Web3, func_prototype: str) -> str:
        result = w3.eth.call(
            {
                "to": self.address,
                "data": encode_function_calldata(
                    function_prototype=func_prototype,
                    function_arguments=None,
                ),
            }
        )

        try:
            (symbol,) = eth_abi.abi.decode(types=["string"], data=result)
            return cast("str", symbol)
        except DecodingError:
            (symbol,) = eth_abi.abi.decode(types=["bytes32"], data=result)
            return cast("HexBytes", symbol).decode("utf-8", errors="ignore").strip("\x00")

    def get_decimals(self, w3: Web3, func_prototype: str) -> int:
        (result,) = raw_call(
            w3=w3,
            address=self.address,
            calldata=encode_function_calldata(
                function_prototype=func_prototype,
                function_arguments=None,
            ),
            return_types=["uint256"],
        )
        return cast("int", result)

    def get_approval(
        self,
        owner: str,
        spender: str,
        block_identifier: BlockIdentifier | None = None,
    ) -> int:
        """
        Retrieve the amount that can be spent by `spender` on behalf of `owner`.
        """

        owner = get_checksum_address(owner)
        spender = get_checksum_address(spender)

        block_number = (
            block_identifier
            if isinstance(block_identifier, int)
            else get_number_for_block_identifier(
                block_identifier,
                self.w3,
            )
        )

        with contextlib.suppress(KeyError):
            return self._cached_approval[block_number, owner, spender]

        approval: int
        (approval,) = eth_abi.abi.decode(
            types=["uint256"],
            data=self.w3.eth.call(
                transaction={
                    "to": self.address,
                    "data": Web3.keccak(text="allowance(address,address)")[:4]
                    + eth_abi.abi.encode(types=["address", "address"], args=[owner, spender]),
                },
                block_identifier=block_number,
            ),
        )
        self._cached_approval[block_number, owner, spender] = approval
        return approval

    async def get_approval_async(
        self,
        owner: str,
        spender: str,
        block_identifier: BlockIdentifier | None = None,
    ) -> int:
        """
        Retrieve the amount that can be spent by `spender` on behalf of `owner`.
        """

        owner = get_checksum_address(owner)
        spender = get_checksum_address(spender)

        block_number = (
            block_identifier
            if isinstance(block_identifier, int)
            else await get_number_for_block_identifier_async(
                block_identifier,
                self.async_w3,
            )
        )

        with contextlib.suppress(KeyError):
            return self._cached_approval[block_number, owner, spender]

        approval: int
        (approval,) = eth_abi.abi.decode(
            types=["uint256"],
            data=await self.async_w3.eth.call(
                transaction={
                    "to": self.address,
                    "data": Web3.keccak(text="allowance(address,address)")[:4]
                    + eth_abi.abi.encode(types=["address", "address"], args=[owner, spender]),
                },
                block_identifier=block_number,
            ),
        )
        self._cached_approval[block_number, owner, spender] = approval
        return approval

    def get_balance(
        self,
        address: str,
        block_identifier: BlockIdentifier | None = None,
    ) -> int:
        """
        Retrieve the ERC-20 balance for the given address.
        """

        address = get_checksum_address(address)

        block_number = (
            block_identifier
            if isinstance(block_identifier, int)
            else get_number_for_block_identifier(
                block_identifier,
                self.w3,
            )
        )

        with contextlib.suppress(KeyError):
            return self._cached_balance[address][block_number]

        balance: int
        (balance,) = eth_abi.abi.decode(
            types=["uint256"],
            data=self.w3.eth.call(
                transaction={
                    "to": self.address,
                    "data": Web3.keccak(text="balanceOf(address)")[:4]
                    + eth_abi.abi.encode(types=["address"], args=[address]),
                },
                block_identifier=block_number,
            ),
        )

        if address not in self._cached_balance:
            self._cached_balance[address] = BoundedCache(max_items=self._state_cache_depth)

        self._cached_balance[address][block_number] = balance
        return balance

    async def get_balance_async(
        self,
        address: str,
        block_identifier: BlockIdentifier | None = None,
    ) -> int:
        """
        Retrieve the ERC-20 balance for the given address.
        """

        address = get_checksum_address(address)

        block_number = (
            block_identifier
            if isinstance(block_identifier, int)
            else await get_number_for_block_identifier_async(
                block_identifier,
                self.async_w3,
            )
        )

        with contextlib.suppress(KeyError):
            return self._cached_balance[address][block_number]

        balance: int
        (balance,) = eth_abi.abi.decode(
            types=["uint256"],
            data=await self.async_w3.eth.call(
                transaction={
                    "to": self.address,
                    "data": Web3.keccak(text="balanceOf(address)")[:4]
                    + eth_abi.abi.encode(types=["address"], args=[address]),
                },
                block_identifier=block_number,
            ),
        )

        if address not in self._cached_balance:
            self._cached_balance[address] = BoundedCache(max_items=self._state_cache_depth)

        self._cached_balance[address][block_number] = balance
        return balance

    def get_total_supply(self, block_identifier: BlockIdentifier | None = None) -> int:
        """
        Retrieve the total supply for this token.
        """

        block_number = (
            block_identifier
            if isinstance(block_identifier, int)
            else get_number_for_block_identifier(
                block_identifier,
                self.w3,
            )
        )

        with contextlib.suppress(KeyError):
            return self._cached_total_supply[block_number]

        total_supply: int
        (total_supply,) = eth_abi.abi.decode(
            types=["uint256"],
            data=self.w3.eth.call(
                transaction={
                    "to": self.address,
                    "data": Web3.keccak(text="totalSupply()")[:4],
                },
                block_identifier=block_number,
            ),
        )
        self._cached_total_supply[block_number] = total_supply
        return total_supply

    async def get_total_supply_async(self, block_identifier: BlockIdentifier | None = None) -> int:
        """
        Retrieve the total supply for this token.
        """

        block_number = (
            block_identifier
            if isinstance(block_identifier, int)
            else await get_number_for_block_identifier_async(
                block_identifier,
                self.async_w3,
            )
        )

        with contextlib.suppress(KeyError):
            return self._cached_total_supply[block_number]

        total_supply: int
        (total_supply,) = eth_abi.abi.decode(
            types=["uint256"],
            data=await self.async_w3.eth.call(
                transaction={
                    "to": self.address,
                    "data": Web3.keccak(text="totalSupply()")[:4],
                },
                block_identifier=block_number,
            ),
        )
        self._cached_total_supply[block_number] = total_supply
        return total_supply

    @property
    def chain_id(self) -> int:
        return self._chain_id

    @property
    def price(self) -> float:
        if self._price_oracle is None:
            raise NoPriceOracle
        return self._price_oracle.price

    @property
    def w3(self) -> Web3:
        return connection_manager.get_web3(self.chain_id)

    @property
    def async_w3(self) -> AsyncWeb3:
        return async_connection_manager.get_web3(self.chain_id)



from . import abi
from .managers import SushiswapV2PoolManager, SushiswapV3PoolManager
from .pools import SushiswapV2Pool, SushiswapV3Pool

__all__ = (
    "SushiswapV2Pool",
    "SushiswapV2PoolManager",
    "SushiswapV3Pool",
    "SushiswapV3PoolManager",
    "abi",
)



from degenbot.balancer.libraries import log_exp_math
from degenbot.balancer.libraries.constants import FOUR, MAX_POW_RELATIVE_ERROR, ONE, TWO
from degenbot.constants import MAX_UINT256
from degenbot.exceptions.evm import EVMRevertError

_ZERO = 0


def add(a: int, b: int) -> int:
    if a + b > MAX_UINT256:
        raise EVMRevertError(error="ADD_OVERFLOW")
    return a + b


def sub(a: int, b: int) -> int:
    if b > a:
        raise EVMRevertError(error="SUB_OVERFLOW")
    return a - b


def mul_down(a: int, b: int) -> int:
    product = a * b
    if not (a == 0 or product // a == b):
        raise EVMRevertError(error="MUL_OVERFLOW")
    return product // ONE


def mul_up(a: int, b: int) -> int:
    product = a * b

    if product == 0:
        return _ZERO
    if product > MAX_UINT256:
        raise EVMRevertError(error="MUL_OVERFLOW")

    return (product - 1) // ONE + 1


def div_down(a: int, b: int) -> int:
    if b == 0:
        raise EVMRevertError(error="ZERO_DIVISION")

    if a == 0:
        return _ZERO

    a_inflated = a * ONE
    if a_inflated > MAX_UINT256:
        raise EVMRevertError(error="DIV_INTERNAL")

    return a_inflated // b


def div_up(a: int, b: int) -> int:
    if b == 0:
        raise EVMRevertError(error="ZERO_DIVISION")

    if a == 0:
        return _ZERO

    a_inflated = a * ONE
    if a_inflated > MAX_UINT256:
        raise EVMRevertError(error="DIV_INTERNAL")

    return ((a_inflated - 1) // b) + 1


def pow_down(x: int, y: int) -> int:
    """
    Returns x^y, assuming both are fixed point numbers, rounding down. The result is guaranteed to
    not be above the true value (that is, the error function expected - actual is always positive).
    """

    # Optimize for when y equals 1.0, 2.0 or 4.0, as those are very simple to implement and occur
    # often in 50/50 and 80/20 Weighted Pools
    if y == ONE:
        return x
    if y == TWO:
        return mul_down(x, x)
    if y == FOUR:
        square = mul_down(x, x)
        return mul_down(square, square)

    raw = log_exp_math.pow(x, y)
    max_error = add(mul_up(raw, MAX_POW_RELATIVE_ERROR), 1)
    if raw < max_error:
        return 0

    return sub(raw, max_error)


def pow_up(x: int, y: int) -> int:
    """
    Returns x^y, assuming both are fixed point numbers, rounding up. The result is guaranteed to not
    be below the true value (that is, the error function expected - actual is always negative).
    """

    # Optimize for when y equals 1.0, 2.0 or 4.0, as those are very simple to implement and occur
    # often in 50/50 and 80/20 Weighted Pools
    if y == ONE:
        return x
    if y == TWO:
        return mul_up(x, x)
    if y == FOUR:
        square = mul_up(x, x)
        return mul_up(square, square)

    raw = log_exp_math.pow(x, y)
    max_error = add(mul_up(raw, MAX_POW_RELATIVE_ERROR), 1)
    return add(raw, max_error)


def complement(x: int) -> int:
    return ONE - x if x < ONE else _ZERO



from .types import ArbitrageCalculationResult
from .uniswap_curve_cycle import UniswapCurveCycle
from .uniswap_lp_cycle import UniswapLpCycle

__all__ = (
    "ArbitrageCalculationResult",
    "UniswapCurveCycle",
    "UniswapLpCycle",
)



from degenbot.constants import MAX_INT128, MAX_UINT128, MIN_INT128, MIN_UINT128
from degenbot.exceptions.evm import EVMRevertError


def add_delta(x: int, y: int) -> int:
    """
    This function has been heavily modified to directly check that the result
    fits in a uint128, instead of checking via < or >= tricks via Solidity's
    built-in casting as implemented at https://github.com/Uniswap/v3-core/blob/main/contracts/libraries/LiquidityMath.sol
    """

    if not (MIN_UINT128 <= x <= MAX_UINT128):
        raise EVMRevertError(error="x not a valid uint128")
    if not (MIN_INT128 <= y <= MAX_INT128):
        raise EVMRevertError(error="y not a valid int128")

    z = x + y

    if y < 0 and not (MIN_UINT128 <= z <= MAX_UINT128):
        raise EVMRevertError(error="LS")
    if not (MIN_UINT128 <= z <= MAX_UINT128):
        raise EVMRevertError(error="LA")

    return z



from collections import OrderedDict, defaultdict
from collections.abc import Callable
from typing import Any, Protocol, Self
from weakref import WeakSet

from degenbot.types.abstract import AbstractPoolState


class KeyedDefaultDict[KT, VT](defaultdict[KT, VT]):
    """
    A modified defaultdict that passes the key to default_factory.
    """

    def __init__(self, default_factory: Callable[[KT], VT]) -> None:
        self._default_factory = default_factory

    def __missing__(self, key: KT) -> VT:
        value = self._default_factory(key)
        self[key] = value
        return value


class AbstractPublisherMessage:
    """
    A message sent by a `Publisher` to a `Subscriber`.
    """


class PoolStateMessage(AbstractPublisherMessage):
    """
    A message notifying that the publisher (a liquidity pool) has updated its state.
    """

    state: AbstractPoolState


class TextMessage(AbstractPublisherMessage):
    """
    A generic text message.
    """

    def __init__(self, text: str) -> None:
        self.text = text

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TextMessage):
            return NotImplemented
        return self.text == other.text

    def __hash__(self) -> int:
        return hash(self.text)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(text={self.text})"

    def __str__(self) -> str:
        return self.text


class Publisher(Protocol):
    """
    Can send a `Message` to a `Subscriber`
    """

    _subscribers: WeakSet["Subscriber"]

    def subscribe(self, subscriber: "Subscriber") -> None:
        """
        Subscribe to receive messages from this `Publisher`
        """

    def unsubscribe(self, subscriber: "Subscriber") -> None:
        """
        Stop receiving messages from this `Publisher`
        """


class PublisherMixin:
    """
    A set of default methods to accept subscribe & unsubscribe requests. Classes using this mixin
    meet the `Publisher` protocol requirements.
    """

    def subscribe(self: Publisher, subscriber: "Subscriber") -> None:
        self._subscribers.add(subscriber)

    def unsubscribe(self: Publisher, subscriber: "Subscriber") -> None:
        self._subscribers.discard(subscriber)


class Subscriber(Protocol):
    """
    Can subscribe to messages from a `Publisher`
    """

    def notify(self, publisher: "Publisher", message: AbstractPublisherMessage) -> None:
        """
        Deliver `message` to `Subscriber`
        """


class BoundedCache[KT, VT](OrderedDict[KT, VT]):
    """
    A cache holding key-value pairs, tracked by entry order. The cache automatically removes old
    items if the number of items would exceed the maximum number of entries set by `max_items`.

    Setting a value at an existing key will overwrite that value without affecting ordering.
    """

    def __init__(self, max_items: int) -> None:
        super().__init__()
        self.max_items = max_items

    def __reduce__(self) -> tuple[Any, ...]:
        state = super().__reduce__()
        return (
            state[0],
            (self.max_items,),  # max_items argument must be provided to properly unpickle
            None,
            None,
            state[4],
        )

    def __setitem__(self, key: KT, value: VT) -> None:
        super().__setitem__(key, value)
        if len(self) > self.max_items:
            self.popitem(last=False)

    def copy(self) -> Self:
        new_copy = self.__class__(max_items=self.max_items)
        for k, v in self.items():
            new_copy[k] = v
        return new_copy



from eth_typing import ChecksumAddress

from degenbot import get_checksum_address
from degenbot.erc20 import Erc20Token
from degenbot.logging import logger


class SimulationLedger:
    """
    A dictionary-like class for tracking token balances across addresses.

    Token balances are organized first by the holding address, then by the
    token contract address.
    """

    def __init__(self) -> None:
        # Entries are recorded as a dict-of-dicts, keyed by address, then by
        # token address
        self.balances: dict[
            ChecksumAddress,  # address holding balance
            dict[
                ChecksumAddress,  # token address
                int,  # balance
            ],
        ] = {}

    def adjust(
        self,
        address: ChecksumAddress | str,
        token: Erc20Token | ChecksumAddress | str,
        amount: int,
    ) -> None:
        """
        Apply an adjustment to the balance for a token held by an address.

        The amount can be positive (credit) or negative (debit). The method
        checksums all addresses prior to use.

        Parameters
        ----------
        address: str | ChecksumAddress
            The address holding the token balance.
        token: Erc20Token | str | ChecksumAddress
            The token being held. May be passed as an address or an ``Erc20Token``
        amount: int
            The amount to adjust. May be negative or positive.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If inputs did not match the expected types.
        """

        _token_address: ChecksumAddress
        if isinstance(token, Erc20Token):
            _token_address = token.address
        else:
            _token_address = get_checksum_address(token)

        _address = get_checksum_address(address)

        address_balance: dict[ChecksumAddress, int]
        try:
            address_balance = self.balances[_address]
        except KeyError:
            address_balance = {}
            self.balances[_address] = address_balance

        logger.debug(f"BALANCE: {_address} {'+' if amount > 0 else ''}{amount} {_token_address}")

        try:
            address_balance[_token_address]
        except KeyError:
            address_balance[_token_address] = 0
        finally:
            address_balance[_token_address] += amount
            if address_balance[_token_address] == 0:
                del address_balance[_token_address]
            if not address_balance:
                del self.balances[_address]

    def token_balance(
        self,
        address: ChecksumAddress | str,
        token: Erc20Token | ChecksumAddress | str,
    ) -> int:
        """
        Get the balance for a given address and token.

        The method checksums all addresses prior to use.

        Parameters
        ----------
        address: str | ChecksumAddress
            The address holding the token balance.
        token: Erc20Token | str | ChecksumAddress
            The token being held. May be passed as an address or an ``Erc20Token``

        Returns
        -------
        int
            The balance of ``token`` at ``address``

        Raises
        ------
        ValueError
            If inputs did not match the expected types.
        """

        _address = get_checksum_address(address)

        if isinstance(token, Erc20Token):
            _token_address = token.address
        else:
            _token_address = get_checksum_address(token)

        address_balances: dict[ChecksumAddress, int]
        try:
            address_balances = self.balances[_address]
        except KeyError:
            address_balances = {}

        return address_balances.get(_token_address, 0)

    def transfer(
        self,
        token: Erc20Token | ChecksumAddress | str,
        amount: int,
        from_addr: ChecksumAddress | str,
        to_addr: ChecksumAddress | str,
    ) -> None:
        """
        Transfer a balance between addresses.

        The method checksums all addresses prior to use.

        Parameters
        ----------
        token: Erc20Token | str | ChecksumAddress
            The token being held. May be passed as an address or an ``Erc20Token``
        amount: int
            The balance to transfer.
        from_addr: str | ChecksumAddress
            The address holding the token balance.
        to_addr: str | ChecksumAddress
            The address holding the token balance.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If inputs did not match the expected types.
        """

        if isinstance(token, Erc20Token):
            _token_address = token.address
        else:
            _token_address = get_checksum_address(token)

        self.adjust(
            address=from_addr,
            token=_token_address,
            amount=-amount,
        )
        self.adjust(
            address=to_addr,
            token=_token_address,
            amount=amount,
        )



from . import abi
from .curve_stableswap_liquidity_pool import CurveStableswapPool
from .types import (
    CurveStableswapPoolSimulationResult,
    CurveStableswapPoolState,
    CurveStableSwapPoolStateUpdated,
)

__all__ = (
    "CurveStableSwapPoolStateUpdated",
    "CurveStableswapPool",
    "CurveStableswapPoolSimulationResult",
    "CurveStableswapPoolState",
    "abi",
)



import tenacity
import web3

from degenbot.exceptions import DegenbotValueError

from .async_connection_manager import AsyncConnectionManager
from .connection_manager import ConnectionManager


def get_async_web3() -> web3.AsyncWeb3:
    if async_connection_manager.default_chain_id is None:
        raise DegenbotValueError(
            message="A default Web3 instance has not been registered."
        ) from None
    return async_connection_manager.get_web3(chain_id=async_connection_manager.default_chain_id)


async def set_async_web3(
    w3: web3.AsyncWeb3,
    *,
    optimize_middleware: bool = True,
) -> None:
    async_w3_connected_check_with_retry = tenacity.AsyncRetrying(
        stop=tenacity.stop_after_delay(10),
        wait=tenacity.wait_exponential_jitter(),
        retry=tenacity.retry_if_result(lambda result: result is False),
    )
    try:
        await async_w3_connected_check_with_retry(w3.is_connected)
    except tenacity.RetryError as exc:
        raise DegenbotValueError(message="Web3 instance is not connected.") from exc

    await async_connection_manager.register_web3(w3, optimize_middleware=optimize_middleware)
    async_connection_manager.set_default_chain(await w3.eth.chain_id)


def get_web3() -> web3.Web3:
    return connection_manager.get_web3(chain_id=connection_manager.default_chain_id)


def set_web3(
    w3: web3.Web3,
    *,
    optimize_middleware: bool = True,
) -> None:
    w3_connected_check_with_retry = tenacity.Retrying(
        stop=tenacity.stop_after_delay(10),
        wait=tenacity.wait_exponential_jitter(),
        retry=tenacity.retry_if_result(lambda result: result is False),
    )
    try:
        w3_connected_check_with_retry(fn=w3.is_connected)
    except tenacity.RetryError as exc:
        raise DegenbotValueError(message="Web3 instance is not connected.") from exc

    connection_manager.register_web3(w3, optimize_middleware=optimize_middleware)
    connection_manager.set_default_chain(w3.eth.chain_id)


connection_manager = ConnectionManager()
async_connection_manager = AsyncConnectionManager()


__all__ = (
    "async_connection_manager",
    "connection_manager",
    "get_web3",
    "get_web3_async",
    "set_web3",
    "set_web3_async",
)



import functools
from typing import overload

from pydantic import validate_call

from degenbot.constants import MAX_UINT160, MAX_UINT256
from degenbot.exceptions.evm import EVMRevertError
from degenbot.uniswap.v3_libraries._config import V3_LIB_CACHE_SIZE
from degenbot.uniswap.v3_libraries.constants import Q96, Q96_RESOLUTION
from degenbot.uniswap.v3_libraries.full_math import muldiv, muldiv_rounding_up
from degenbot.uniswap.v3_libraries.functions import to_int256, to_uint160
from degenbot.uniswap.v3_libraries.unsafe_math import div_rounding_up
from degenbot.validation.evm_values import (
    ValidatedInt128,
    ValidatedInt256,
    ValidatedUint128,
    ValidatedUint128NonZero,
    ValidatedUint160,
    ValidatedUint160NonZero,
    ValidatedUint256,
)

"""
ref: https://github.com/Uniswap/v3-core/blob/main/contracts/libraries/SqrtPriceMath.sol
"""


@overload
def get_amount0_delta(
    *,
    sqrt_ratio_a_x96: ValidatedUint160,
    sqrt_ratio_b_x96: ValidatedUint160,
    liquidity: ValidatedInt128,
    round_up: bool,
) -> ValidatedUint256: ...


@overload
def get_amount0_delta(
    *,
    sqrt_ratio_a_x96: ValidatedUint160,
    sqrt_ratio_b_x96: ValidatedUint160,
    liquidity: ValidatedInt128,
    round_up: None,
) -> ValidatedInt256: ...


@functools.lru_cache(maxsize=V3_LIB_CACHE_SIZE)
@validate_call(validate_return=True)
def get_amount0_delta(
    *,
    sqrt_ratio_a_x96: ValidatedUint160NonZero,
    sqrt_ratio_b_x96: ValidatedUint160,
    liquidity: ValidatedInt128 | ValidatedUint128,
    round_up: bool | None = None,
) -> ValidatedInt256 | ValidatedUint256:
    # The Solidity function is overloaded with respect to `roundUp`.
    # ref: https://github.com/Uniswap/v3-core/blob/main/contracts/libraries/sqrt_price_math.sol

    if round_up is not None:
        if sqrt_ratio_a_x96 > sqrt_ratio_b_x96:
            sqrt_ratio_a_x96, sqrt_ratio_b_x96 = sqrt_ratio_b_x96, sqrt_ratio_a_x96

        numerator1 = liquidity << Q96_RESOLUTION
        numerator2 = sqrt_ratio_b_x96 - sqrt_ratio_a_x96

        return (
            div_rounding_up(
                muldiv_rounding_up(numerator1, numerator2, sqrt_ratio_b_x96),
                sqrt_ratio_a_x96,
            )
            if round_up
            else muldiv(numerator1, numerator2, sqrt_ratio_b_x96) // sqrt_ratio_a_x96
        )

    return to_int256(
        to_int256(
            -get_amount0_delta(
                sqrt_ratio_a_x96=sqrt_ratio_a_x96,
                sqrt_ratio_b_x96=sqrt_ratio_b_x96,
                liquidity=-liquidity,
                round_up=False,
            )
        )
        if liquidity < 0
        else to_int256(
            get_amount0_delta(
                sqrt_ratio_a_x96=sqrt_ratio_a_x96,
                sqrt_ratio_b_x96=sqrt_ratio_b_x96,
                liquidity=liquidity,
                round_up=True,
            )
        )
    )


@overload
def get_amount1_delta(
    *,
    sqrt_ratio_a_x96: ValidatedUint160,
    sqrt_ratio_b_x96: ValidatedUint160,
    liquidity: ValidatedInt128,
    round_up: bool,
) -> ValidatedUint256: ...


@overload
def get_amount1_delta(
    *,
    sqrt_ratio_a_x96: ValidatedUint160,
    sqrt_ratio_b_x96: ValidatedUint160,
    liquidity: ValidatedInt128,
    round_up: None,
) -> ValidatedInt256: ...


@functools.lru_cache(maxsize=V3_LIB_CACHE_SIZE)
@validate_call(validate_return=True)
def get_amount1_delta(
    *,
    sqrt_ratio_a_x96: ValidatedUint160,
    sqrt_ratio_b_x96: ValidatedUint160,
    liquidity: ValidatedInt128 | ValidatedUint128,
    round_up: bool | None = None,
) -> ValidatedInt256 | ValidatedUint256:
    # The Solidity function is overloaded with respect to `roundUp`. Both modes are encapsulated
    # here by the optional `round_up` argument.

    assert liquidity >= 0

    if round_up is not None:
        if sqrt_ratio_a_x96 > sqrt_ratio_b_x96:
            sqrt_ratio_a_x96, sqrt_ratio_b_x96 = sqrt_ratio_b_x96, sqrt_ratio_a_x96

        return (
            muldiv_rounding_up(liquidity, sqrt_ratio_b_x96 - sqrt_ratio_a_x96, Q96)
            if round_up
            else muldiv(liquidity, sqrt_ratio_b_x96 - sqrt_ratio_a_x96, Q96)
        )

    return to_int256(
        to_int256(
            -get_amount1_delta(
                sqrt_ratio_a_x96=sqrt_ratio_a_x96,
                sqrt_ratio_b_x96=sqrt_ratio_b_x96,
                liquidity=-liquidity,
                round_up=False,
            )
        )
        if liquidity < 0
        else to_int256(
            get_amount1_delta(
                sqrt_ratio_a_x96=sqrt_ratio_a_x96,
                sqrt_ratio_b_x96=sqrt_ratio_b_x96,
                liquidity=liquidity,
                round_up=True,
            )
        )
    )


@functools.lru_cache(maxsize=V3_LIB_CACHE_SIZE)
@validate_call(validate_return=True)
def get_next_sqrt_price_from_amount0_rounding_up(
    *,
    sqrt_price_x96: ValidatedUint160,
    liquidity: ValidatedUint128,
    amount: ValidatedUint256,
    add: bool,
) -> ValidatedUint160:
    if amount == 0:
        return sqrt_price_x96

    numerator1 = liquidity << Q96_RESOLUTION
    product = amount * sqrt_price_x96

    if add:
        if product < MAX_UINT256:  # safe path, no overflow
            denominator = numerator1 + product
            if denominator >= numerator1:
                return muldiv_rounding_up(
                    a=numerator1,
                    b=sqrt_price_x96,
                    denominator=denominator,
                )
        # failsafe path in case of overflow
        return div_rounding_up(
            x=numerator1,
            y=(numerator1 // sqrt_price_x96) + amount,
        )

    if not numerator1 > product:
        raise EVMRevertError(error="required: numerator1 > product")
    denominator = numerator1 - product
    return to_uint160(
        muldiv_rounding_up(
            a=numerator1,
            b=sqrt_price_x96,
            denominator=denominator,
        )
    )


@functools.lru_cache(maxsize=V3_LIB_CACHE_SIZE)
@validate_call(validate_return=True)
def get_next_sqrt_price_from_amount1_rounding_down(
    *,
    sqrt_price_x96: ValidatedUint160,
    liquidity: ValidatedUint128,
    amount: ValidatedUint256,
    add: bool,
) -> ValidatedUint160:
    if add:
        quotient = (
            (amount << Q96_RESOLUTION) // liquidity
            if amount <= MAX_UINT160
            else muldiv(amount, Q96, liquidity)
        )
        return to_uint160(sqrt_price_x96 + quotient)

    quotient = (
        div_rounding_up(amount << Q96_RESOLUTION, liquidity)
        if amount <= MAX_UINT160
        else muldiv_rounding_up(amount, Q96, liquidity)
    )

    if not (sqrt_price_x96 > quotient):
        raise EVMRevertError(error="require sqrtPX96 > quotient")

    # always fits 160 bits
    return sqrt_price_x96 - quotient


@functools.lru_cache(maxsize=V3_LIB_CACHE_SIZE)
@validate_call(validate_return=True)
def get_next_sqrt_price_from_input(
    *,
    sqrt_price_x96: ValidatedUint160NonZero,
    liquidity: ValidatedUint128NonZero,
    amount_in: ValidatedUint256,
    zero_for_one: bool,
) -> ValidatedUint160:
    if not (sqrt_price_x96 > 0):
        raise EVMRevertError(error="required: sqrt_price_x96 > 0")

    if not (liquidity > 0):
        raise EVMRevertError(error="required: liquidity > 0")

    # round to make sure that we don't pass the target price
    return (
        get_next_sqrt_price_from_amount0_rounding_up(
            sqrt_price_x96=sqrt_price_x96,
            liquidity=liquidity,
            amount=amount_in,
            add=True,
        )
        if zero_for_one
        else get_next_sqrt_price_from_amount1_rounding_down(
            sqrt_price_x96=sqrt_price_x96,
            liquidity=liquidity,
            amount=amount_in,
            add=True,
        )
    )


@functools.lru_cache(maxsize=V3_LIB_CACHE_SIZE)
@validate_call(validate_return=True)
def get_next_sqrt_price_from_output(
    *,
    sqrt_price_x96: ValidatedUint160NonZero,
    liquidity: ValidatedUint128NonZero,
    amount_out: ValidatedUint256,
    zero_for_one: bool,
) -> int:
    if not (sqrt_price_x96 > 0):
        raise EVMRevertError(error="required: sqrt_price_x96 > 0")

    if not (liquidity > 0):
        raise EVMRevertError(error="required: liquidity must be > 0")

    # round to make sure that we pass the target price
    return (
        get_next_sqrt_price_from_amount1_rounding_down(
            sqrt_price_x96=sqrt_price_x96,
            liquidity=liquidity,
            amount=amount_out,
            add=False,
        )
        if zero_for_one
        else get_next_sqrt_price_from_amount0_rounding_up(
            sqrt_price_x96=sqrt_price_x96,
            liquidity=liquidity,
            amount=amount_out,
            add=False,
        )
    )



from fractions import Fraction

from degenbot import connection_manager, get_checksum_address
from degenbot.camelot.functions import get_y_camelot, k_camelot
from degenbot.erc20 import Erc20Token
from degenbot.exceptions import DegenbotValueError
from degenbot.functions import encode_function_calldata, raw_call
from degenbot.logging import logger
from degenbot.types.aliases import ChainId
from degenbot.uniswap.v2_liquidity_pool import UniswapV2Pool
from degenbot.uniswap.v2_types import UniswapV2PoolState


class CamelotLiquidityPool(UniswapV2Pool):
    CAMELOT_ARBITRUM_POOL_INIT_HASH = (
        "0xa856464ae65f7619087bc369daaf7e387dae1e5af69cfa7935850ebf754b04c1"
    )

    def __init__(
        self,
        address: str,
        *,
        chain_id: ChainId | None = None,
        silent: bool = False,
    ) -> None:
        address = get_checksum_address(address)

        if chain_id is None:  # pragma: no branch
            chain_id = connection_manager.default_chain_id

        w3 = connection_manager.get_web3(chain_id)
        state_block = w3.eth.get_block_number()

        fee_token0: int
        fee_token1: int
        _, _, fee_token0, fee_token1 = raw_call(
            w3=w3,
            address=address,
            calldata=encode_function_calldata(
                function_prototype="getReserves()",
                function_arguments=None,
            ),
            return_types=["uint256", "uint256", "uint256", "uint256"],
        )

        self.fee_denominator: int
        (self.fee_denominator,) = raw_call(
            w3=w3,
            address=address,
            calldata=encode_function_calldata(
                function_prototype="FEE_DENOMINATOR()",
                function_arguments=None,
            ),
            return_types=["uint256"],
        )

        super().__init__(
            address=address,
            chain_id=chain_id,
            init_hash=self.CAMELOT_ARBITRUM_POOL_INIT_HASH,
            fee=(
                Fraction(fee_token0, self.fee_denominator),
                Fraction(fee_token1, self.fee_denominator),
            ),
            silent=silent,
            state_block=state_block,
        )

        self.stable_swap: bool
        (self.stable_swap,) = raw_call(
            w3=w3,
            address=address,
            calldata=encode_function_calldata(
                function_prototype="stableSwap()",
                function_arguments=None,
            ),
            return_types=["bool"],
        )

    def _calculate_tokens_out_from_tokens_in_stable_swap(
        self,
        token_in: Erc20Token,
        token_in_quantity: int,
        override_state: UniswapV2PoolState | None = None,
    ) -> int:
        """
        Calculates the expected token OUTPUT for a target INPUT at current pool reserves.
        Uses the self.token0 and self.token1 pointers to determine which token is being swapped in
        """

        if override_state is not None:  # pragma: no cover
            logger.debug(f"State overrides applied: {override_state}")

        if token_in_quantity <= 0:  # pragma: no cover
            raise DegenbotValueError(message="token_in_quantity must be positive")

        precision_multiplier_token0: int = 10**self.token0.decimals
        precision_multiplier_token1: int = 10**self.token1.decimals

        fee_percent = self.fee_denominator * (
            self.fee_token0 if token_in == self.token0 else self.fee_token1
        )

        reserves_token0 = (
            override_state.reserves_token0 if override_state is not None else self.reserves_token0
        )
        reserves_token1 = (
            override_state.reserves_token1 if override_state is not None else self.reserves_token1
        )

        # Remove fee from amount received
        token_in_quantity -= token_in_quantity * fee_percent // self.fee_denominator
        xy = k_camelot(
            balance_0=reserves_token0,
            balance_1=reserves_token1,
            decimals_0=precision_multiplier_token0,
            decimals_1=precision_multiplier_token1,
        )
        reserves_token0 = reserves_token0 * 10**18 // precision_multiplier_token0
        reserves_token1 = reserves_token1 * 10**18 // precision_multiplier_token1
        reserve_a, reserve_b = (
            (reserves_token0, reserves_token1)
            if token_in == self.token0
            else (reserves_token1, reserves_token0)
        )
        token_in_quantity = (
            token_in_quantity * 10**18 // precision_multiplier_token0
            if token_in == self.token0
            else token_in_quantity * 10**18 // precision_multiplier_token1
        )
        y = reserve_b - get_y_camelot(token_in_quantity + reserve_a, xy, reserve_b)

        return (
            y
            * (
                precision_multiplier_token1
                if token_in == self.token0
                else precision_multiplier_token0
            )
            // 10**18
        )

    def calculate_tokens_out_from_tokens_in(
        self,
        token_in: Erc20Token,
        token_in_quantity: int,
        override_state: UniswapV2PoolState | None = None,
    ) -> int:
        if self.stable_swap:
            return self._calculate_tokens_out_from_tokens_in_stable_swap(
                token_in=token_in,
                token_in_quantity=token_in_quantity,
                override_state=override_state,
            )
        return super().calculate_tokens_out_from_tokens_in(
            token_in=token_in,
            token_in_quantity=token_in_quantity,
            override_state=override_state,
        )



import asyncio
import math
from collections.abc import Iterable, Mapping, Sequence
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from dataclasses import dataclass
from fractions import Fraction
from typing import Any
from weakref import WeakSet

import eth_abi.abi
from eth_typing import ChecksumAddress, HexStr
from scipy.optimize import OptimizeResult, minimize_scalar
from web3 import Web3

from degenbot import get_checksum_address
from degenbot.aerodrome import AerodromeV2Pool, AerodromeV2PoolState, AerodromeV3Pool
from degenbot.arbitrage.types import (
    ArbitrageCalculationResult,
    UniswapV2PoolSwapAmounts,
    UniswapV3PoolSwapAmounts,
    UniswapV4PoolSwapAmounts,
)
from degenbot.constants import WRAPPED_NATIVE_TOKENS, ZERO_ADDRESS
from degenbot.erc20 import Erc20Token, EtherPlaceholder
from degenbot.exceptions import DegenbotValueError
from degenbot.exceptions.evm import EVMRevertError
from degenbot.exceptions.arbitrage import ArbitrageError, RateOfExchangeBelowMinimum
from degenbot.exceptions.liquidity_pool import LiquidityPoolError
from degenbot.logging import logger
from degenbot.types.abstract import AbstractArbitrage
from degenbot.types.concrete import (
    AbstractPublisherMessage,
    PoolStateMessage,
    Publisher,
    PublisherMixin,
    Subscriber,
    TextMessage,
)
from degenbot.uniswap.types import UniswapPoolSwapVector
from degenbot.uniswap.v2_liquidity_pool import UniswapV2Pool
from degenbot.uniswap.v2_types import UniswapV2PoolState
from degenbot.uniswap.v3_libraries.tick_math import MAX_SQRT_RATIO, MIN_SQRT_RATIO
from degenbot.uniswap.v3_liquidity_pool import UniswapV3Pool
from degenbot.uniswap.v3_types import UniswapV3PoolState
from degenbot.uniswap.v4_liquidity_pool import UniswapV4Pool
from degenbot.uniswap.v4_types import UniswapV4PoolState

type Pool = AerodromeV2Pool | AerodromeV3Pool | UniswapV2Pool | UniswapV3Pool | UniswapV4Pool
type PoolState = AerodromeV2PoolState | UniswapV2PoolState | UniswapV3PoolState | UniswapV4PoolState
type SwapAmount = UniswapV2PoolSwapAmounts | UniswapV3PoolSwapAmounts | UniswapV4PoolSwapAmounts
type PoolId = bytes | HexStr


@dataclass(slots=True, frozen=True)
class V4PoolKey:
    currency0: ChecksumAddress
    currency1: ChecksumAddress
    fee: int
    tick_spacing: int
    hooks: ChecksumAddress


UNISWAP_V2_SWAP_FUNCTION_SELECTOR = Web3.keccak(text="swap(uint256,uint256,address,bytes)")[:4]
UNISWAP_V3_SWAP_FUNCTION_SELECTOR = Web3.keccak(text="swap(address,bool,int256,uint160,bytes)")[:4]
ERC20_TOKEN_TRANSFER_FUNCTION_SELECTOR = Web3.keccak(text="transfer(address,uint256)")[:4]


class UniswapLpCycle(PublisherMixin, AbstractArbitrage):
    def _notify_subscribers(self: Publisher, message: AbstractPublisherMessage) -> None:
        for subscriber in self._subscribers:
            subscriber.notify(publisher=self, message=message)

    def __init__(
        self,
        input_token: Erc20Token,
        swap_pools: Iterable[Pool],
        id: str,  # noqa:A002
        max_input: int | None = None,
    ) -> None:
        for swap_pool in swap_pools:
            if not isinstance(swap_pool, Pool.__value__):
                raise DegenbotValueError(
                    message=f"Incompatible pool type ({type(swap_pool)}) provided."
                )
        self.swap_pools: tuple[Pool, ...] = tuple(swap_pools)
        if len(set(self.swap_pools)) != len(self.swap_pools):
            raise DegenbotValueError(message="Swap pools must not contain duplicates.")

        self.id = id
        self.input_token = input_token

        if max_input is None:
            logger.warning("No maximum input provided, setting to 100 WETH")
            max_input = 100 * 10**18
        elif max_input <= 0:
            raise DegenbotValueError(message="Maximum input must be positive.")
        self.max_input = max_input

        _swap_vectors: list[UniswapPoolSwapVector] = []
        for i, pool in enumerate(self.swap_pools):
            input_token = self.input_token if i == 0 else _swap_vectors[-1].token_out

            match input_token:
                case pool.token0:
                    _swap_vectors.append(
                        UniswapPoolSwapVector(
                            token_in=pool.token0,
                            token_out=pool.token1,
                            zero_for_one=pool.token0 == input_token,
                        )
                    )
                case pool.token1:
                    _swap_vectors.append(
                        UniswapPoolSwapVector(
                            token_in=pool.token1,
                            token_out=pool.token0,
                            zero_for_one=pool.token0 == input_token,
                        )
                    )
                case EtherPlaceholder() if (
                    wrapped_native_token := WRAPPED_NATIVE_TOKENS[pool.chain_id]
                ) in pool.tokens:
                    # Handle case where input token is Ether and pool holds wrapped native
                    _swap_vectors.append(
                        UniswapPoolSwapVector(
                            token_in=pool.token0
                            if pool.token0 == wrapped_native_token
                            else pool.token1,
                            token_out=pool.token1
                            if pool.token0 == wrapped_native_token
                            else pool.token0,
                            zero_for_one=pool.token0 == wrapped_native_token,
                        )
                    )
                case Erc20Token() if (
                    input_token == WRAPPED_NATIVE_TOKENS[pool.chain_id]
                    and ZERO_ADDRESS in pool.tokens
                ):
                    # Handle case where input token is wrapped native and pool holds Ether
                    _swap_vectors.append(
                        UniswapPoolSwapVector(
                            token_in=pool.token0 if pool.token0 == ZERO_ADDRESS else pool.token1,
                            token_out=pool.token1 if pool.token0 == ZERO_ADDRESS else pool.token0,
                            zero_for_one=pool.token1 == ZERO_ADDRESS,
                        )
                    )
                case _:
                    raise DegenbotValueError(
                        message=f"Token {input_token} could not be matched. Pool holds {pool.token0} & {pool.token1}"  # noqa: E501
                    )

        self._swap_vectors = tuple(_swap_vectors)

        self.name = " → ".join([pool.name for pool in self.swap_pools])

        self._pool_viability: dict[Pool, bool] = {
            pool: self._pool_is_viable(pool=pool, state=pool.state, vector=swap_vector)
            for pool, swap_vector in zip(self.swap_pools, self._swap_vectors, strict=True)
        }
        assert len(self._pool_viability) == len(self.swap_pools), f"{self.id=} {self.swap_pools=}"

        self._subscribers: WeakSet[Subscriber] = WeakSet()
        for pool in self.swap_pools:
            pool.subscribe(self)

    def __getstate__(self) -> dict[str, Any]:
        dropped_attributes = ("_subscribers",)
        copied_attributes = ()

        return {
            k: (v.copy() if k in copied_attributes else v)
            for k, v in self.__dict__.items()
            if k not in dropped_attributes
        }

    def __str__(self) -> str:
        return self.name

    def _build_swap_amounts(
        self,
        token_in_quantity: int,
        state_overrides: Mapping[Pool, PoolState] | None = None,
    ) -> tuple[SwapAmount, ...]:
        """
        Generate inputs for all swaps along the arbitrage path, starting with the specified amount
        of the input token defined in the constructor.
        """

        if state_overrides is None:
            state_overrides = {}

        _token_out_quantity = 0
        swap_amounts: list[SwapAmount] = []
        for i, (pool, swap_vector) in enumerate(
            zip(self.swap_pools, self._swap_vectors, strict=True)
        ):
            _token_in_quantity = token_in_quantity if i == 0 else _token_out_quantity

            if _token_in_quantity == 0:
                raise ArbitrageError(message="Zero amount swap")

            try:
                match pool, state_overrides.get(pool):
                    case AerodromeV2Pool(), (
                        AerodromeV2PoolState() | None
                    ) as aerodrome_v2_pool_state:
                        _token_out_quantity = pool.calculate_tokens_out_from_tokens_in(
                            token_in=swap_vector.token_in,
                            token_in_quantity=_token_in_quantity,
                            override_state=aerodrome_v2_pool_state,
                        )
                        if _token_out_quantity == 0:  # pragma: no cover
                            raise ArbitrageError(
                                message=f"Zero-output swap through pool {pool} @ {pool.address}"
                            )
                        swap_amounts.append(
                            UniswapV2PoolSwapAmounts(
                                pool=pool.address,
                                amounts_in=(_token_in_quantity, 0)
                                if swap_vector.zero_for_one
                                else (0, _token_in_quantity),
                                amounts_out=(0, _token_out_quantity)
                                if swap_vector.zero_for_one
                                else (_token_out_quantity, 0),
                            )
                        )
                    case UniswapV2Pool(), (UniswapV2PoolState() | None) as uniswap_v2_pool_state:
                        _token_out_quantity = pool.calculate_tokens_out_from_tokens_in(
                            token_in=swap_vector.token_in,
                            token_in_quantity=_token_in_quantity,
                            override_state=uniswap_v2_pool_state,
                        )
                        if _token_out_quantity == 0:  # pragma: no cover
                            raise ArbitrageError(
                                message=f"Zero-output swap through pool {pool} @ {pool.address}"
                            )
                        swap_amounts.append(
                            UniswapV2PoolSwapAmounts(
                                pool=pool.address,
                                amounts_in=(_token_in_quantity, 0)
                                if swap_vector.zero_for_one
                                else (0, _token_in_quantity),
                                amounts_out=(0, _token_out_quantity)
                                if swap_vector.zero_for_one
                                else (_token_out_quantity, 0),
                            )
                        )
                    case UniswapV3Pool(), (UniswapV3PoolState() | None) as uniswap_v3_pool_state:
                        _token_out_quantity = pool.calculate_tokens_out_from_tokens_in(
                            token_in=swap_vector.token_in,
                            token_in_quantity=_token_in_quantity,
                            override_state=uniswap_v3_pool_state,
                        )
                        if _token_out_quantity == 0:  # pragma: no cover
                            raise ArbitrageError(
                                message=f"Zero-output swap through pool {pool} @ {pool.address}"
                            )
                        swap_amounts.append(
                            UniswapV3PoolSwapAmounts(
                                pool=pool.address,
                                amount_specified=_token_in_quantity,
                                zero_for_one=swap_vector.zero_for_one,
                                sqrt_price_limit_x96=MIN_SQRT_RATIO + 1
                                if swap_vector.zero_for_one
                                else MAX_SQRT_RATIO - 1,
                            )
                        )
                    case UniswapV4Pool(), (UniswapV4PoolState() | None) as uniswap_v4_pool_state:
                        _token_out_quantity = pool.calculate_tokens_out_from_tokens_in(
                            token_in=swap_vector.token_in,
                            token_in_quantity=_token_in_quantity,
                            override_state=uniswap_v4_pool_state,
                        )
                        if _token_out_quantity == 0:  # pragma: no cover
                            raise ArbitrageError(
                                message=f"Zero-output swap through pool {pool} @ {pool.address}"
                            )
                        swap_amounts.append(
                            UniswapV4PoolSwapAmounts(
                                address=pool.address,
                                id=pool.pool_id,
                                amount_specified=_token_in_quantity,
                                zero_for_one=swap_vector.zero_for_one,
                                sqrt_price_limit_x96=MIN_SQRT_RATIO + 1
                                if swap_vector.zero_for_one
                                else MAX_SQRT_RATIO - 1,
                            )
                        )
                    case _:  # pragma: no cover
                        raise DegenbotValueError(
                            message="Could not identify pool and override type."
                        )
            except LiquidityPoolError as exc:  # pragma: no cover
                raise ArbitrageError(message=str(exc)) from exc

        return tuple(swap_amounts)

    def _pool_is_viable(
        self,
        pool: Pool,
        state: PoolState,
        vector: UniswapPoolSwapVector,
    ) -> bool:
        """
        Check if the pool can perform a swap along the given vector.
        """

        match pool, state:
            case AerodromeV2Pool(), AerodromeV2PoolState():
                return pool.swap_is_viable(state=state, vector=vector)
            case UniswapV2Pool(), UniswapV2PoolState():
                return pool.swap_is_viable(state=state, vector=vector)
            case UniswapV3Pool(), UniswapV3PoolState():
                return pool.swap_is_viable(state=state, vector=vector)
            case UniswapV4Pool(), UniswapV4PoolState():
                return pool.swap_is_viable(state=state, vector=vector)
            case _:  # pragma: no cover
                raise DegenbotValueError(
                    message=f"Could not identify pool {pool} and state {state}."
                )

    def _pre_calculation_check(
        self,
        min_rate_of_exchange: Fraction = Fraction(1, 1),
        state_overrides: Mapping[Pool, PoolState] | None = None,
    ) -> None:
        """
        Perform pool viability and minimum rate of exchange checks. Raises an exception if a
        non-viable pool is found, or if the instantaneous rate of exchange is below the specified
        minimum.
        """

        if state_overrides is None:
            state_overrides = {}

        # Check the pools for viability
        for pool, vector in zip(self.swap_pools, self._swap_vectors, strict=True):
            assert pool in self._pool_viability

            if pool in state_overrides:
                if not self._pool_is_viable(pool=pool, state=state_overrides[pool], vector=vector):
                    raise ArbitrageError(message=f"Pool {pool!r} is not viable")
            elif self._pool_viability[pool] is False:
                raise ArbitrageError(message=f"Pool {pool!r} is not viable")

        # Evaluate the instantaneous rate of exchange for the path by accumulating swap and fee
        # multipliers for each pool
        exchange_rate_multipliers: list[Fraction] = []
        for pool, vector in zip(self.swap_pools, self._swap_vectors, strict=True):
            match pool, state_overrides.get(pool):
                case AerodromeV2Pool(), (AerodromeV2PoolState() | None) as aerodrome_v2_pool_state:
                    # The multiplier for the pool is the rate of exchange for the output token,
                    # reduced by the fee taken on the input amount
                    swap_multiplier = pool.get_absolute_exchange_rate(
                        token=vector.token_out,
                        override_state=aerodrome_v2_pool_state,
                    )
                    fee = pool.fee_token0 if vector.zero_for_one else pool.fee_token1
                    fee_multiplier = Fraction(fee.denominator - fee.numerator, fee.denominator)

                    exchange_rate_multipliers.append(swap_multiplier)
                    exchange_rate_multipliers.append(fee_multiplier)

                case UniswapV2Pool(), (UniswapV2PoolState() | None) as uniswap_v2_pool_state:
                    # The multiplier for the pool is the rate of exchange for the output token,
                    # reduced by the fee taken on the input amount
                    swap_multiplier = pool.get_absolute_exchange_rate(
                        token=vector.token_out,
                        override_state=uniswap_v2_pool_state,
                    )
                    fee = pool.fee_token0 if vector.zero_for_one else pool.fee_token1
                    fee_multiplier = Fraction(fee.denominator - fee.numerator, fee.denominator)

                    exchange_rate_multipliers.append(swap_multiplier)
                    exchange_rate_multipliers.append(fee_multiplier)

                case UniswapV3Pool(), (UniswapV3PoolState() | None) as uniswap_v3_pool_state:
                    swap_multiplier = pool.get_absolute_exchange_rate(
                        token=vector.token_out,
                        override_state=uniswap_v3_pool_state,
                    )

                    # TODO: add support for mixed-fee pools if necessary
                    fee = Fraction(pool.fee, pool.FEE_DENOMINATOR)
                    fee_multiplier = Fraction(fee.denominator - fee.numerator, fee.denominator)

                    exchange_rate_multipliers.append(swap_multiplier)
                    exchange_rate_multipliers.append(fee_multiplier)

                case UniswapV4Pool(), (UniswapV4PoolState() | None) as uniswap_v4_pool_state:
                    swap_multiplier = pool.get_absolute_exchange_rate(
                        token=vector.token_out,
                        override_state=uniswap_v4_pool_state,
                    )

                    # TODO: add support for mixed-fee pools if necessary
                    fee = Fraction(pool.fee, pool.FEE_DENOMINATOR)
                    fee_multiplier = Fraction(fee.denominator - fee.numerator, fee.denominator)

                    exchange_rate_multipliers.append(swap_multiplier)
                    exchange_rate_multipliers.append(fee_multiplier)

                case _:  # pragma: no cover
                    raise DegenbotValueError(message=f"Could not identify pool {pool}.")

        net_rate_of_exchange = Fraction(
            math.prod(multiplier.numerator for multiplier in exchange_rate_multipliers),
            math.prod(multiplier.denominator for multiplier in exchange_rate_multipliers),
        )
        if net_rate_of_exchange < min_rate_of_exchange:
            raise RateOfExchangeBelowMinimum(net_rate_of_exchange)

    def _calculate(
        self,
        state_overrides: Mapping[Pool, PoolState] | None = None,
    ) -> ArbitrageCalculationResult:
        """
        Calculate the optimal arbitrage profit using the maximum input as an upper bound.
        """

        if state_overrides is None:
            state_overrides = {}

        # The bounded Brent optimizer requires bounds for the input amount, and a bracketed guess
        # to initiate the search
        bounds: tuple[float, float] = (
            1.0,
            float(self.max_input),
        )
        bracket: tuple[float, float] = (0.25 * self.max_input, 0.50 * self.max_input)

        def arb_profit(x: float) -> float:
            token_in_quantity = int(x)  # round the input down
            token_out_quantity: int = 0

            try:
                for i, (pool, swap_vector) in enumerate(
                    zip(self.swap_pools, self._swap_vectors, strict=True)
                ):
                    state = state_overrides.get(pool)

                    match pool, state:
                        case AerodromeV2Pool(), (
                            AerodromeV2PoolState() | None
                        ) as aerodrome_v2_pool_state:
                            token_out_quantity = pool.calculate_tokens_out_from_tokens_in(
                                token_in=swap_vector.token_in,
                                token_in_quantity=token_in_quantity
                                if i == 0
                                else token_out_quantity,
                                override_state=aerodrome_v2_pool_state,
                            )
                        case UniswapV2Pool(), (
                            UniswapV2PoolState() | None
                        ) as uniswap_v2_pool_state:
                            token_out_quantity = pool.calculate_tokens_out_from_tokens_in(
                                token_in=swap_vector.token_in,
                                token_in_quantity=token_in_quantity
                                if i == 0
                                else token_out_quantity,
                                override_state=uniswap_v2_pool_state,
                            )
                        case UniswapV3Pool(), (
                            UniswapV3PoolState() | None
                        ) as uniswap_v3_pool_state:
                            token_out_quantity = pool.calculate_tokens_out_from_tokens_in(
                                token_in=swap_vector.token_in,
                                token_in_quantity=token_in_quantity
                                if i == 0
                                else token_out_quantity,
                                override_state=uniswap_v3_pool_state,
                            )
                        case UniswapV4Pool(), (
                            UniswapV4PoolState() | None
                        ) as uniswap_v4_pool_state:
                            token_out_quantity = pool.calculate_tokens_out_from_tokens_in(
                                token_in=swap_vector.token_in,
                                token_in_quantity=token_in_quantity
                                if i == 0
                                else token_out_quantity,
                                override_state=uniswap_v4_pool_state,
                            )
                        case _:  # pragma: no cover
                            raise DegenbotValueError(
                                message=f"Unrecognized pool {pool} and state {state}."
                            )
            except (EVMRevertError, LiquidityPoolError):  # pragma: no cover
                # The optimizer might send invalid amounts into the swap calculation during
                # iteration. We don't want it to stop, so catch the exception and pretend the
                # swap resulted in zero output
                token_out_quantity = 0

            # minimize_scalar requires the function to have a minimum value
            # for the solver to settle on an optimum input, so return the
            # negated profit
            return float(token_out_quantity - token_in_quantity)

        opt: OptimizeResult = minimize_scalar(
            fun=lambda x: -arb_profit(x),
            method="bounded",
            bounds=bounds,
            bracket=bracket,
            options={"xatol": 1.0},
        )

        # Negate the result to convert to a sensible value (positive profit)
        best_profit = -int(opt.fun)
        swap_amount = int(opt.x)

        best_amounts = self._build_swap_amounts(
            token_in_quantity=swap_amount,
            state_overrides=state_overrides,
        )

        newest_state_block = None
        if not state_overrides:
            pool_state_blocks = tuple(
                block for pool in self.swap_pools if (block := pool.state.block) is not None
            )
            if len(pool_state_blocks) == len(self.swap_pools):
                newest_state_block = max(pool_state_blocks)

        return ArbitrageCalculationResult(
            id=self.id,
            input_token=self.input_token,
            profit_token=self.input_token,
            input_amount=swap_amount,
            profit_amount=best_profit,
            swap_amounts=best_amounts,
            state_block=newest_state_block,
        )

    def calculate(
        self,
        state_overrides: Mapping[Pool, PoolState] | None = None,
        min_rate_of_exchange: Fraction = Fraction(1, 1),
    ) -> ArbitrageCalculationResult:
        """
        Calculate the results of the arbitrage at the current pool states, or at one or more
        overridden pool states if provided.
        """

        self._pre_calculation_check(
            min_rate_of_exchange=min_rate_of_exchange,
            state_overrides=state_overrides,
        )

        return self._calculate(state_overrides=state_overrides)

    async def calculate_with_pool(
        self,
        executor: ProcessPoolExecutor | ThreadPoolExecutor,
        state_overrides: Mapping[Pool, PoolState] | None = None,
        min_rate_of_exchange: Fraction = Fraction(1, 1),
    ) -> asyncio.Future[ArbitrageCalculationResult]:
        """
        Wrap the arbitrage calculation into an asyncio future using the specified executor.

        Arguments
        ---------
        executor : Executor
            An executor (from `concurrent.futures`) to process the calculation work. Both
            `ThreadPoolExecutor` and `ProcessPoolExecutor` are supported, but `ProcessPoolExecutor`
            is recommended for CPU-bound work like this.
        state_overrides : Mapping[ChecksumAddress, UniswapPoolStateOverride], optional
            A dict (or equivalent mapping) of pool states, keyed by the checksummed address of that
            pool.
        min_rate_of_exchange : Fraction, optional
            The minimum net rate of exchange for the arbitrage path. Rates below this minimum will
            raise an exception.

        Returns
        -------
        A future which returns a `ArbitrageCalculationResult` (or exception) when awaited.

        Notes
        -----
        This is an async function that must be called with the `await` keyword.
        """

        if isinstance(executor, ProcessPoolExecutor) and any(
            pool.sparse_liquidity_map for pool in self.swap_pools if isinstance(pool, UniswapV3Pool)
        ):
            raise DegenbotValueError(
                message="Cannot perform calculation with process pool executor. One or more V3 pools has a sparse bitmap."  # noqa:E501
            )

        self._pre_calculation_check(
            min_rate_of_exchange=min_rate_of_exchange,
            state_overrides=state_overrides,
        )

        return asyncio.get_running_loop().run_in_executor(
            executor,
            self._calculate,
            state_overrides,
        )

    def generate_payloads(
        self,
        from_address: ChecksumAddress | str,
        swap_amount: int,
        pool_swap_amounts: Sequence[SwapAmount],
    ) -> list[tuple[ChecksumAddress, bytes, int]]:
        """
        Generate a list of ABI-encoded calldata for each step in the swap path.

        Calldata is built using the eth_abi.encode method and the ABI for the
        ``swap`` function at the Uniswap pool. V2 and V3 pools are supported.

        Arguments
        ---------
        from_address: str
            The address that will execute the calldata. Must be a smart
            contract implementing the required callbacks specific to the pool.

        swap_amount: int
            The initial amount of `token_in` to swap through the first pool.

        pool_swap_amounts: Iterable[UniswapV2PoolSwapAmounts |
        UniswapV3PoolSwapAmounts] | UniswapV4PoolSwapAmounts]
            An iterable of swap amounts to be encoded.

        Returns
        -------
        ``list[(str, bytes, int)]``
            A list of payload tuples. Each payload tuple has form (
            address: ChecksumAddress, calldata: bytes, value: int).
        """

        from_address = get_checksum_address(from_address)

        msg_value = 0  # This arbitrage does not require a `msg.value` payment
        payloads = []
        for i, (swap_pool, _swap_amounts) in enumerate(
            zip(self.swap_pools, pool_swap_amounts, strict=True)
        ):
            # Special case when a Uniswap V2 pool is the next step in the path
            if i < len(self.swap_pools) - 1 and isinstance(
                (next_pool := self.swap_pools[i + 1]), AerodromeV2Pool | UniswapV2Pool
            ):
                swap_destination_address = next_pool.address
            else:
                swap_destination_address = from_address

            match swap_pool, _swap_amounts:
                case AerodromeV2Pool() | UniswapV2Pool(), UniswapV2PoolSwapAmounts():
                    # Special case: If first pool is type V2, input token must be transferred prior
                    # to the swap
                    if i == 0:
                        payloads.append(
                            (
                                # address
                                self.input_token.address,
                                # bytes calldata
                                ERC20_TOKEN_TRANSFER_FUNCTION_SELECTOR
                                + eth_abi.abi.encode(
                                    types=(
                                        "address",
                                        "uint256",
                                    ),
                                    args=(
                                        swap_pool.address,
                                        swap_amount,
                                    ),
                                ),
                                msg_value,
                            )
                        )

                    logger.debug(f"PAYLOAD: building V2 swap at pool {i}")
                    logger.debug(f"PAYLOAD: pool address {swap_pool.address}")
                    logger.debug(f"PAYLOAD: swap amounts {_swap_amounts}")
                    logger.debug(f"PAYLOAD: destination address {swap_destination_address}")

                    payloads.append(
                        (
                            # address
                            swap_pool.address,
                            # bytes calldata
                            UNISWAP_V2_SWAP_FUNCTION_SELECTOR
                            + eth_abi.abi.encode(
                                types=(
                                    "uint256",
                                    "uint256",
                                    "address",
                                    "bytes",
                                ),
                                args=(
                                    *_swap_amounts.amounts_out,
                                    swap_destination_address,
                                    b"",
                                ),
                            ),
                            msg_value,
                        )
                    )
                case UniswapV3Pool(), UniswapV3PoolSwapAmounts():
                    logger.debug(f"PAYLOAD: building V3 swap at pool {i}")
                    logger.debug(f"PAYLOAD: pool address {swap_pool.address}")
                    logger.debug(f"PAYLOAD: swap amounts {_swap_amounts}")
                    logger.debug(f"PAYLOAD: destination address {swap_destination_address}")

                    payloads.append(
                        (
                            # address
                            swap_pool.address,
                            # bytes calldata
                            UNISWAP_V3_SWAP_FUNCTION_SELECTOR
                            + eth_abi.abi.encode(
                                types=(
                                    "address",
                                    "bool",
                                    "int256",
                                    "uint160",
                                    "bytes",
                                ),
                                args=(
                                    swap_destination_address,
                                    _swap_amounts.zero_for_one,
                                    _swap_amounts.amount_specified,
                                    _swap_amounts.sqrt_price_limit_x96,
                                    b"",
                                ),
                            ),
                            msg_value,
                        )
                    )
                case UniswapV4Pool(), UniswapV4PoolSwapAmounts():
                    logger.debug(f"PAYLOAD: building V4 swap at pool {i}")
                    logger.debug(f"PAYLOAD: pool address {swap_pool.address}")
                    logger.debug(f"PAYLOAD: swap amounts {_swap_amounts}")
                    logger.debug(f"PAYLOAD: destination address {swap_destination_address}")

                    payloads.append(
                        V4PoolKey(
                            currency0=swap_pool.token0.address,
                            currency1=swap_pool.token1.address,
                            fee=swap_pool.fee,
                            tick_spacing=swap_pool.tick_spacing,
                            hooks=swap_pool.hook_address,
                        )
                    )

                case _:  # pragma: no cover
                    raise DegenbotValueError(message="Could not identify pool and swap amounts.")

        return payloads

    def notify(self, publisher: Publisher, message: AbstractPublisherMessage) -> None:
        match publisher, message:
            case (
                (
                    AerodromeV2Pool()
                    | AerodromeV3Pool()
                    | UniswapV2Pool()
                    | UniswapV3Pool()
                    | UniswapV4Pool()
                ) as pool,
                PoolStateMessage() as state_message,
            ) if pool in self.swap_pools and isinstance(
                state_message.state,
                (
                    AerodromeV2PoolState
                    | UniswapV2PoolState
                    | UniswapV3PoolState
                    | UniswapV4PoolState
                ),
            ):
                # Check the pool's viability at the new state
                self._pool_viability[pool] = self._pool_is_viable(
                    pool=pool,
                    state=state_message.state,
                    vector=self._swap_vectors[self.swap_pools.index(pool)],
                )

                try:
                    self._pre_calculation_check()
                except ArbitrageError:
                    return
                else:
                    self._notify_subscribers(TextMessage("Profitable state discovered."))

            case _:  # pragma: no cover
                logger.error(
                    f"Message {message} from publisher {publisher} was not handled by {self}"
                )




from pydantic import validate_call

from degenbot.validation.evm_values import ValidatedInt128, ValidatedUint128


@validate_call(validate_return=True)
def add_delta(
    x: ValidatedUint128,
    y: ValidatedInt128,
) -> ValidatedUint128:
    """
    This function has been modified to check that the result fits in a uint128, instead
    of inline Yul as implemented by the Solidity contract.

    ref: https://github.com/Uniswap/v4-core/blob/main/src/libraries/LiquidityMath.sol
    """

    return x + y



import pytest
import web3

from degenbot import (
    AnvilFork,
    async_connection_manager,
    connection_manager,
    get_web3,
    set_async_web3,
    set_web3,
)
from degenbot.exceptions import DegenbotValueError

from .conftest import ETHEREUM_ARCHIVE_NODE_HTTP_URI


def test_disconnected_web3():
    w3 = web3.Web3(web3.HTTPProvider("https://google.com"))
    with pytest.raises(DegenbotValueError, match="Web3 instance is not connected."):
        set_web3(w3)

    with pytest.raises(DegenbotValueError, match="Web3 instance is not connected."):
        connection_manager.register_web3(w3)


def test_legacy_interface(fork_mainnet_full: AnvilFork):
    with pytest.raises(
        DegenbotValueError, match="A default Web3 instance has not been registered."
    ):
        get_web3()

    set_web3(fork_mainnet_full.w3)
    assert get_web3() is fork_mainnet_full.w3


def test_optimized_web3():
    w3 = web3.Web3(web3.HTTPProvider(ETHEREUM_ARCHIVE_NODE_HTTP_URI))
    middlewares = w3.middleware_onion.middleware
    set_web3(w3)
    assert w3.middleware_onion.middleware == []

    w3 = web3.Web3(web3.HTTPProvider(ETHEREUM_ARCHIVE_NODE_HTTP_URI))
    middlewares = w3.middleware_onion.middleware
    set_web3(w3, optimize_middleware=False)
    assert w3.middleware_onion.middleware == middlewares


def test_connection_manager(fork_mainnet_full: AnvilFork):
    with pytest.raises(DegenbotValueError):
        _ = connection_manager.default_chain_id

    set_web3(fork_mainnet_full.w3)
    assert connection_manager.default_chain_id == fork_mainnet_full.w3.eth.chain_id
    assert connection_manager.get_web3(fork_mainnet_full.w3.eth.chain_id) is fork_mainnet_full.w3

    with pytest.raises(DegenbotValueError):
        connection_manager.get_web3(69)


async def test_async_connection_manager(fork_mainnet_full: AnvilFork):
    async with fork_mainnet_full.async_w3 as async_w3:
        await set_async_web3(async_w3)
        assert async_connection_manager.default_chain_id == await async_w3.eth.chain_id
        assert async_connection_manager.get_web3(await async_w3.eth.chain_id) is async_w3



import pytest
from eth_typing import Hash32, HexStr
from eth_utils.crypto import keccak
from hexbytes import HexBytes

from degenbot import AnvilFork, get_checksum_address, set_web3
from degenbot.exceptions import DegenbotValueError
from degenbot.functions import (
    create2_address,
    encode_function_calldata,
    extract_argument_types_from_function_prototype,
    get_number_for_block_identifier,
    next_base_fee,
    raw_call,
)
from degenbot.types.aliases import BlockNumber


def test_extract_argument_types_from_function_prototype():
    assert extract_argument_types_from_function_prototype("func()") == []
    assert extract_argument_types_from_function_prototype("func(uint256)") == [
        "uint256",
    ]
    assert extract_argument_types_from_function_prototype("func(uint256,address)") == [
        "uint256",
        "address",
    ]
    assert extract_argument_types_from_function_prototype("func(uint256,address,bytes)") == [
        "uint256",
        "address",
        "bytes",
    ]
    assert extract_argument_types_from_function_prototype("func(uint256,address,bytes[])") == [
        "uint256",
        "address",
        "bytes[]",
    ]


def test_encode_function_calldata():
    assert (
        encode_function_calldata(function_prototype="factory()", function_arguments=[])
        == HexBytes("0xc45a01550ceb4bc5c6b2e6f722b5033a03078f9bd6673457375ba94c26ac1cf0")[:4]
    )
    assert encode_function_calldata(
        function_prototype="transfer(address,uint256)",
        function_arguments=[
            "0xA69babEF1cA67A37Ffaf7a485DfFF3382056e78C",
            26535330612692929974,
        ],
    ) == HexBytes(
        "0xa9059cbb000000000000000000000000a69babef1ca67a37ffaf7a485dfff3382056e78c00000000000000000000000000000000000000000000000170406e9a1f1c4db6"
    )


def test_low_level_call_for_factory_address(fork_mainnet_full: AnvilFork):
    set_web3(fork_mainnet_full.w3)

    pool_address = get_checksum_address("0xCBCdF9626bC03E24f779434178A73a0B4bad62eD")

    function_prototype = "factory()"

    (result,) = raw_call(
        w3=fork_mainnet_full.w3,
        block_identifier=fork_mainnet_full.w3.eth.block_number,
        address=pool_address,
        calldata=encode_function_calldata(
            function_prototype=function_prototype,
            function_arguments=extract_argument_types_from_function_prototype(function_prototype),
        ),
        return_types=["address"],
    )
    assert get_checksum_address(result) == get_checksum_address(
        "0x1F98431c8aD98523631AE4a59f267346ea31F984"
    )


def test_create2():
    """
    Tests taken from https://eips.ethereum.org/EIPS/eip-1014
    """

    assert (
        create2_address(
            deployer="0x0000000000000000000000000000000000000000",
            salt="0x0000000000000000000000000000000000000000000000000000000000000000",
            init_code_hash=keccak(hexstr="0x00"),
        )
        == "0x4D1A2e2bB4F88F0250f26Ffff098B0b30B26BF38"
    )
    assert (
        create2_address(
            deployer="0xdeadbeef00000000000000000000000000000000",
            salt="0x0000000000000000000000000000000000000000000000000000000000000000",
            init_code_hash=keccak(hexstr="0x00"),
        )
        == "0xB928f69Bb1D91Cd65274e3c79d8986362984fDA3"
    )
    assert (
        create2_address(
            deployer="0xdeadbeef00000000000000000000000000000000",
            salt="0x000000000000000000000000feed000000000000000000000000000000000000",
            init_code_hash=keccak(hexstr="0x00"),
        )
        == "0xD04116cDd17beBE565EB2422F2497E06cC1C9833"
    )
    assert (
        create2_address(
            deployer="0x0000000000000000000000000000000000000000",
            salt="0x0000000000000000000000000000000000000000000000000000000000000000",
            init_code_hash=keccak(hexstr="0xdeadbeef"),
        )
        == "0x70f2b2914A2a4b783FaEFb75f459A580616Fcb5e"
    )
    assert (
        create2_address(
            deployer="0x00000000000000000000000000000000deadbeef",
            salt="0x00000000000000000000000000000000000000000000000000000000cafebabe",
            init_code_hash=keccak(hexstr="0xdeadbeef"),
        )
        == "0x60f3f640a8508fC6a86d45DF051962668E1e8AC7"
    )
    assert (
        create2_address(
            deployer="0x00000000000000000000000000000000deadbeef",
            salt="0x00000000000000000000000000000000000000000000000000000000cafebabe",
            init_code_hash=keccak(
                hexstr="0xdeadbeefdeadbeefdeadbeefdeadbeefdeadbeefdeadbeefdeadbeefdeadbeefdeadbeefdeadbeefdeadbeef"
            ),
        )
        == "0x1d8bfDC5D46DC4f61D6b6115972536eBE6A8854C"
    )


def test_converting_block_identifier_to_int(fork_mainnet_full: AnvilFork):
    """
    Check that all inputs for web3 type `BlockIdentifier` can be converted to an integer
    """

    w3 = fork_mainnet_full.w3
    set_web3(w3)

    # Known string literals
    latest_block = get_number_for_block_identifier("latest", w3)
    earliest_block = get_number_for_block_identifier("earliest", w3)
    pending_block = get_number_for_block_identifier("pending", w3)
    safe_block = get_number_for_block_identifier("safe", w3)
    finalized_block = get_number_for_block_identifier("finalized", w3)

    assert BlockNumber.__value__ is int
    assert isinstance(latest_block, BlockNumber.__value__)
    assert isinstance(earliest_block, BlockNumber.__value__)
    assert isinstance(pending_block, BlockNumber.__value__)
    assert isinstance(safe_block, BlockNumber.__value__)
    assert isinstance(finalized_block, BlockNumber.__value__)

    assert latest_block != earliest_block != pending_block != safe_block != finalized_block

    # BlockNumber
    assert isinstance(
        get_number_for_block_identifier(1, w3),
        BlockNumber.__value__,
    )

    # Hash32
    assert isinstance(
        get_number_for_block_identifier(Hash32((1).to_bytes(length=32, byteorder="big")), w3),
        BlockNumber.__value__,
    )

    # HexStr
    assert isinstance(
        get_number_for_block_identifier(
            HexStr("0x" + (128).to_bytes(32, byteorder="big").hex()), w3
        ),
        BlockNumber.__value__,
    )

    # HexBytes
    assert isinstance(get_number_for_block_identifier(HexBytes(1), w3), BlockNumber.__value__)

    # int
    assert isinstance(get_number_for_block_identifier(1, w3), BlockNumber.__value__)

    for invalid_tag in ["Latest", "latest ", "next", "previous"]:
        with pytest.raises(DegenbotValueError):
            get_number_for_block_identifier(invalid_tag, w3)  # type: ignore[arg-type]

    with pytest.raises(DegenbotValueError):
        get_number_for_block_identifier(1.0, w3)  # type: ignore[arg-type]


def test_fee_calcs():
    base_fee = 100 * 10**9

    # EIP-1559 target is 50% full blocks, so a 50% full block should return the same base fee
    assert (
        next_base_fee(
            parent_base_fee=base_fee,
            parent_gas_used=15_000_000,
            parent_gas_limit=30_000_000,
        )
        == base_fee
    )

    # Fee should be higher
    assert (
        next_base_fee(
            parent_base_fee=base_fee,
            parent_gas_used=20_000_000,
            parent_gas_limit=30_000_000,
        )
        == 104166666666
    )

    # Fee should be lower
    assert (
        next_base_fee(
            parent_base_fee=base_fee,
            parent_gas_used=10_000_000,
            parent_gas_limit=30_000_000,
        )
        == 95833333334
    )

    min_base_fee = 95 * 10**9

    # Enforce minimum fee
    assert (
        next_base_fee(
            parent_base_fee=base_fee,
            parent_gas_used=0,
            parent_gas_limit=30_000_000,
            min_base_fee=min_base_fee,
        )
        == min_base_fee
    )
